def get_swin_model_order(version: int) -> list[str]:
    if version == 2:
        order = [
            "[0].features.layers[0].layers[0].weight",
            "[0].features.layers[0].layers[0].bias",
            "[0].features.layers[0].layers[2].weight",
            "[0].features.layers[0].layers[2].bias",
            "[0].features.layers[1].layers[0].norm1.weight",
            "[0].features.layers[1].layers[0].norm1.bias",
            "[0].features.layers[1].layers[0].attn.logit_scale",
            "[1][<flat index 1>]",
            "[1][<flat index 0>]",
            "[0].features.layers[1].layers[0].attn.qkv.weight",
            "[0].features.layers[1].layers[0].attn.qkv.bias",
            "[0].features.layers[1].layers[0].attn.proj.weight",
            "[0].features.layers[1].layers[0].attn.proj.bias",
            "[0].features.layers[1].layers[0].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[1].layers[0].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[1].layers[0].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[1].layers[0].norm2.weight",
            "[0].features.layers[1].layers[0].norm2.bias",
            "[0].features.layers[1].layers[0].mlp.layers[0].weight",
            "[0].features.layers[1].layers[0].mlp.layers[0].bias",
            "[0].features.layers[1].layers[0].mlp.layers[3].weight",
            "[0].features.layers[1].layers[0].mlp.layers[3].bias",
            "[0].features.layers[1].layers[1].norm1.weight",
            "[0].features.layers[1].layers[1].norm1.bias",
            "[0].features.layers[1].layers[1].attn.logit_scale",
            "[1][<flat index 3>]",
            "[1][<flat index 2>]",
            "[0].features.layers[1].layers[1].attn.qkv.weight",
            "[0].features.layers[1].layers[1].attn.qkv.bias",
            "[0].features.layers[1].layers[1].attn.proj.weight",
            "[0].features.layers[1].layers[1].attn.proj.bias",
            "[0].features.layers[1].layers[1].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[1].layers[1].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[1].layers[1].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[1].layers[1].norm2.weight",
            "[0].features.layers[1].layers[1].norm2.bias",
            "[0].features.layers[1].layers[1].mlp.layers[0].weight",
            "[0].features.layers[1].layers[1].mlp.layers[0].bias",
            "[0].features.layers[1].layers[1].mlp.layers[3].weight",
            "[0].features.layers[1].layers[1].mlp.layers[3].bias",
            "[0].features.layers[2].reduction.weight",
            "[0].features.layers[2].norm.weight",
            "[0].features.layers[2].norm.bias",
            "[0].features.layers[3].layers[0].norm1.weight",
            "[0].features.layers[3].layers[0].norm1.bias",
            "[0].features.layers[3].layers[0].attn.logit_scale",
            "[1][<flat index 5>]",
            "[1][<flat index 4>]",
            "[0].features.layers[3].layers[0].attn.qkv.weight",
            "[0].features.layers[3].layers[0].attn.qkv.bias",
            "[0].features.layers[3].layers[0].attn.proj.weight",
            "[0].features.layers[3].layers[0].attn.proj.bias",
            "[0].features.layers[3].layers[0].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[3].layers[0].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[3].layers[0].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[3].layers[0].norm2.weight",
            "[0].features.layers[3].layers[0].norm2.bias",
            "[0].features.layers[3].layers[0].mlp.layers[0].weight",
            "[0].features.layers[3].layers[0].mlp.layers[0].bias",
            "[0].features.layers[3].layers[0].mlp.layers[3].weight",
            "[0].features.layers[3].layers[0].mlp.layers[3].bias",
            "[0].features.layers[3].layers[1].norm1.weight",
            "[0].features.layers[3].layers[1].norm1.bias",
            "[0].features.layers[3].layers[1].attn.logit_scale",
            "[1][<flat index 7>]",
            "[1][<flat index 6>]",
            "[0].features.layers[3].layers[1].attn.qkv.weight",
            "[0].features.layers[3].layers[1].attn.qkv.bias",
            "[0].features.layers[3].layers[1].attn.proj.weight",
            "[0].features.layers[3].layers[1].attn.proj.bias",
            "[0].features.layers[3].layers[1].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[3].layers[1].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[3].layers[1].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[3].layers[1].norm2.weight",
            "[0].features.layers[3].layers[1].norm2.bias",
            "[0].features.layers[3].layers[1].mlp.layers[0].weight",
            "[0].features.layers[3].layers[1].mlp.layers[0].bias",
            "[0].features.layers[3].layers[1].mlp.layers[3].weight",
            "[0].features.layers[3].layers[1].mlp.layers[3].bias",
            "[0].features.layers[4].reduction.weight",
            "[0].features.layers[4].norm.weight",
            "[0].features.layers[4].norm.bias",
            "[0].features.layers[5].layers[0].norm1.weight",
            "[0].features.layers[5].layers[0].norm1.bias",
            "[0].features.layers[5].layers[0].attn.logit_scale",
            "[1][<flat index 9>]",
            "[1][<flat index 8>]",
            "[0].features.layers[5].layers[0].attn.qkv.weight",
            "[0].features.layers[5].layers[0].attn.qkv.bias",
            "[0].features.layers[5].layers[0].attn.proj.weight",
            "[0].features.layers[5].layers[0].attn.proj.bias",
            "[0].features.layers[5].layers[0].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[5].layers[0].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[5].layers[0].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[5].layers[0].norm2.weight",
            "[0].features.layers[5].layers[0].norm2.bias",
            "[0].features.layers[5].layers[0].mlp.layers[0].weight",
            "[0].features.layers[5].layers[0].mlp.layers[0].bias",
            "[0].features.layers[5].layers[0].mlp.layers[3].weight",
            "[0].features.layers[5].layers[0].mlp.layers[3].bias",
            "[0].features.layers[5].layers[1].norm1.weight",
            "[0].features.layers[5].layers[1].norm1.bias",
            "[0].features.layers[5].layers[1].attn.logit_scale",
            "[1][<flat index 11>]",
            "[1][<flat index 10>]",
            "[0].features.layers[5].layers[1].attn.qkv.weight",
            "[0].features.layers[5].layers[1].attn.qkv.bias",
            "[0].features.layers[5].layers[1].attn.proj.weight",
            "[0].features.layers[5].layers[1].attn.proj.bias",
            "[0].features.layers[5].layers[1].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[5].layers[1].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[5].layers[1].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[5].layers[1].norm2.weight",
            "[0].features.layers[5].layers[1].norm2.bias",
            "[0].features.layers[5].layers[1].mlp.layers[0].weight",
            "[0].features.layers[5].layers[1].mlp.layers[0].bias",
            "[0].features.layers[5].layers[1].mlp.layers[3].weight",
            "[0].features.layers[5].layers[1].mlp.layers[3].bias",
            "[0].features.layers[5].layers[2].norm1.weight",
            "[0].features.layers[5].layers[2].norm1.bias",
            "[0].features.layers[5].layers[2].attn.logit_scale",
            "[1][<flat index 13>]",
            "[1][<flat index 12>]",
            "[0].features.layers[5].layers[2].attn.qkv.weight",
            "[0].features.layers[5].layers[2].attn.qkv.bias",
            "[0].features.layers[5].layers[2].attn.proj.weight",
            "[0].features.layers[5].layers[2].attn.proj.bias",
            "[0].features.layers[5].layers[2].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[5].layers[2].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[5].layers[2].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[5].layers[2].norm2.weight",
            "[0].features.layers[5].layers[2].norm2.bias",
            "[0].features.layers[5].layers[2].mlp.layers[0].weight",
            "[0].features.layers[5].layers[2].mlp.layers[0].bias",
            "[0].features.layers[5].layers[2].mlp.layers[3].weight",
            "[0].features.layers[5].layers[2].mlp.layers[3].bias",
            "[0].features.layers[5].layers[3].norm1.weight",
            "[0].features.layers[5].layers[3].norm1.bias",
            "[0].features.layers[5].layers[3].attn.logit_scale",
            "[1][<flat index 15>]",
            "[1][<flat index 14>]",
            "[0].features.layers[5].layers[3].attn.qkv.weight",
            "[0].features.layers[5].layers[3].attn.qkv.bias",
            "[0].features.layers[5].layers[3].attn.proj.weight",
            "[0].features.layers[5].layers[3].attn.proj.bias",
            "[0].features.layers[5].layers[3].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[5].layers[3].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[5].layers[3].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[5].layers[3].norm2.weight",
            "[0].features.layers[5].layers[3].norm2.bias",
            "[0].features.layers[5].layers[3].mlp.layers[0].weight",
            "[0].features.layers[5].layers[3].mlp.layers[0].bias",
            "[0].features.layers[5].layers[3].mlp.layers[3].weight",
            "[0].features.layers[5].layers[3].mlp.layers[3].bias",
            "[0].features.layers[5].layers[4].norm1.weight",
            "[0].features.layers[5].layers[4].norm1.bias",
            "[0].features.layers[5].layers[4].attn.logit_scale",
            "[1][<flat index 17>]",
            "[1][<flat index 16>]",
            "[0].features.layers[5].layers[4].attn.qkv.weight",
            "[0].features.layers[5].layers[4].attn.qkv.bias",
            "[0].features.layers[5].layers[4].attn.proj.weight",
            "[0].features.layers[5].layers[4].attn.proj.bias",
            "[0].features.layers[5].layers[4].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[5].layers[4].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[5].layers[4].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[5].layers[4].norm2.weight",
            "[0].features.layers[5].layers[4].norm2.bias",
            "[0].features.layers[5].layers[4].mlp.layers[0].weight",
            "[0].features.layers[5].layers[4].mlp.layers[0].bias",
            "[0].features.layers[5].layers[4].mlp.layers[3].weight",
            "[0].features.layers[5].layers[4].mlp.layers[3].bias",
            "[0].features.layers[5].layers[5].norm1.weight",
            "[0].features.layers[5].layers[5].norm1.bias",
            "[0].features.layers[5].layers[5].attn.logit_scale",
            "[1][<flat index 19>]",
            "[1][<flat index 18>]",
            "[0].features.layers[5].layers[5].attn.qkv.weight",
            "[0].features.layers[5].layers[5].attn.qkv.bias",
            "[0].features.layers[5].layers[5].attn.proj.weight",
            "[0].features.layers[5].layers[5].attn.proj.bias",
            "[0].features.layers[5].layers[5].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[5].layers[5].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[5].layers[5].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[5].layers[5].norm2.weight",
            "[0].features.layers[5].layers[5].norm2.bias",
            "[0].features.layers[5].layers[5].mlp.layers[0].weight",
            "[0].features.layers[5].layers[5].mlp.layers[0].bias",
            "[0].features.layers[5].layers[5].mlp.layers[3].weight",
            "[0].features.layers[5].layers[5].mlp.layers[3].bias",
            "[0].features.layers[6].reduction.weight",
            "[0].features.layers[6].norm.weight",
            "[0].features.layers[6].norm.bias",
            "[0].features.layers[7].layers[0].norm1.weight",
            "[0].features.layers[7].layers[0].norm1.bias",
            "[0].features.layers[7].layers[0].attn.logit_scale",
            "[1][<flat index 21>]",
            "[1][<flat index 20>]",
            "[0].features.layers[7].layers[0].attn.qkv.weight",
            "[0].features.layers[7].layers[0].attn.qkv.bias",
            "[0].features.layers[7].layers[0].attn.proj.weight",
            "[0].features.layers[7].layers[0].attn.proj.bias",
            "[0].features.layers[7].layers[0].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[7].layers[0].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[7].layers[0].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[7].layers[0].norm2.weight",
            "[0].features.layers[7].layers[0].norm2.bias",
            "[0].features.layers[7].layers[0].mlp.layers[0].weight",
            "[0].features.layers[7].layers[0].mlp.layers[0].bias",
            "[0].features.layers[7].layers[0].mlp.layers[3].weight",
            "[0].features.layers[7].layers[0].mlp.layers[3].bias",
            "[0].features.layers[7].layers[1].norm1.weight",
            "[0].features.layers[7].layers[1].norm1.bias",
            "[0].features.layers[7].layers[1].attn.logit_scale",
            "[1][<flat index 23>]",
            "[1][<flat index 22>]",
            "[0].features.layers[7].layers[1].attn.qkv.weight",
            "[0].features.layers[7].layers[1].attn.qkv.bias",
            "[0].features.layers[7].layers[1].attn.proj.weight",
            "[0].features.layers[7].layers[1].attn.proj.bias",
            "[0].features.layers[7].layers[1].attn.cpb_mlp_ln1.weight",
            "[0].features.layers[7].layers[1].attn.cpb_mlp_ln1.bias",
            "[0].features.layers[7].layers[1].attn.cpb_mlp_ln2.weight",
            "[0].features.layers[7].layers[1].norm2.weight",
            "[0].features.layers[7].layers[1].norm2.bias",
            "[0].features.layers[7].layers[1].mlp.layers[0].weight",
            "[0].features.layers[7].layers[1].mlp.layers[0].bias",
            "[0].features.layers[7].layers[1].mlp.layers[3].weight",
            "[0].features.layers[7].layers[1].mlp.layers[3].bias",
            "[0].norm.weight",
            "[0].norm.bias",
            "[0].head.weight",
            "[0].head.bias",
        ]
    else:
        order = [
            "[0].features.layers[0].layers[0].weight",
            "[0].features.layers[0].layers[0].bias",
            "[0].features.layers[0].layers[2].weight",
            "[0].features.layers[0].layers[2].bias",
            "[0].features.layers[1].layers[0].norm1.weight",
            "[0].features.layers[1].layers[0].norm1.bias",
            "[0].features.layers[1].layers[0].attn.relative_position_bias_table",
            "[1][<flat index 0>]",
            "[0].features.layers[1].layers[0].attn.qkv.weight",
            "[0].features.layers[1].layers[0].attn.qkv.bias",
            "[0].features.layers[1].layers[0].attn.proj.weight",
            "[0].features.layers[1].layers[0].attn.proj.bias",
            "[0].features.layers[1].layers[0].norm2.weight",
            "[0].features.layers[1].layers[0].norm2.bias",
            "[0].features.layers[1].layers[0].mlp.layers[0].weight",
            "[0].features.layers[1].layers[0].mlp.layers[0].bias",
            "[0].features.layers[1].layers[0].mlp.layers[3].weight",
            "[0].features.layers[1].layers[0].mlp.layers[3].bias",
            # Stage 1, Block 1 (features.1.1 -> layers[1].layers[1])
            "[0].features.layers[1].layers[1].norm1.weight",
            "[0].features.layers[1].layers[1].norm1.bias",
            "[0].features.layers[1].layers[1].attn.relative_position_bias_table",
            "[1][<flat index 1>]",
            "[0].features.layers[1].layers[1].attn.qkv.weight",
            "[0].features.layers[1].layers[1].attn.qkv.bias",
            "[0].features.layers[1].layers[1].attn.proj.weight",
            "[0].features.layers[1].layers[1].attn.proj.bias",
            "[0].features.layers[1].layers[1].norm2.weight",
            "[0].features.layers[1].layers[1].norm2.bias",
            "[0].features.layers[1].layers[1].mlp.layers[0].weight",
            "[0].features.layers[1].layers[1].mlp.layers[0].bias",
            "[0].features.layers[1].layers[1].mlp.layers[3].weight",
            "[0].features.layers[1].layers[1].mlp.layers[3].bias",
            "[0].features.layers[2].reduction.weight",
            "[0].features.layers[2].norm.weight",
            "[0].features.layers[2].norm.bias",
            "[0].features.layers[3].layers[0].norm1.weight",
            "[0].features.layers[3].layers[0].norm1.bias",
            "[0].features.layers[3].layers[0].attn.relative_position_bias_table",
            "[1][<flat index 2>]",
            "[0].features.layers[3].layers[0].attn.qkv.weight",
            "[0].features.layers[3].layers[0].attn.qkv.bias",
            "[0].features.layers[3].layers[0].attn.proj.weight",
            "[0].features.layers[3].layers[0].attn.proj.bias",
            "[0].features.layers[3].layers[0].norm2.weight",
            "[0].features.layers[3].layers[0].norm2.bias",
            "[0].features.layers[3].layers[0].mlp.layers[0].weight",
            "[0].features.layers[3].layers[0].mlp.layers[0].bias",
            "[0].features.layers[3].layers[0].mlp.layers[3].weight",
            "[0].features.layers[3].layers[0].mlp.layers[3].bias",
            "[0].features.layers[3].layers[1].norm1.weight",
            "[0].features.layers[3].layers[1].norm1.bias",
            "[0].features.layers[3].layers[1].attn.relative_position_bias_table",
            "[1][<flat index 3>]",
            "[0].features.layers[3].layers[1].attn.qkv.weight",
            "[0].features.layers[3].layers[1].attn.qkv.bias",
            "[0].features.layers[3].layers[1].attn.proj.weight",
            "[0].features.layers[3].layers[1].attn.proj.bias",
            "[0].features.layers[3].layers[1].norm2.weight",
            "[0].features.layers[3].layers[1].norm2.bias",
            "[0].features.layers[3].layers[1].mlp.layers[0].weight",
            "[0].features.layers[3].layers[1].mlp.layers[0].bias",
            "[0].features.layers[3].layers[1].mlp.layers[3].weight",
            "[0].features.layers[3].layers[1].mlp.layers[3].bias",
            "[0].features.layers[4].reduction.weight",
            "[0].features.layers[4].norm.weight",
            "[0].features.layers[4].norm.bias",
            "[0].features.layers[5].layers[0].norm1.weight",
            "[0].features.layers[5].layers[0].norm1.bias",
            "[0].features.layers[5].layers[0].attn.relative_position_bias_table",
            "[1][<flat index 4>]",
            "[0].features.layers[5].layers[0].attn.qkv.weight",
            "[0].features.layers[5].layers[0].attn.qkv.bias",
            "[0].features.layers[5].layers[0].attn.proj.weight",
            "[0].features.layers[5].layers[0].attn.proj.bias",
            "[0].features.layers[5].layers[0].norm2.weight",
            "[0].features.layers[5].layers[0].norm2.bias",
            "[0].features.layers[5].layers[0].mlp.layers[0].weight",
            "[0].features.layers[5].layers[0].mlp.layers[0].bias",
            "[0].features.layers[5].layers[0].mlp.layers[3].weight",
            "[0].features.layers[5].layers[0].mlp.layers[3].bias",
            # Stage 3, Block 1 (features.5.1 -> layers[5].layers[1])
            "[0].features.layers[5].layers[1].norm1.weight",
            "[0].features.layers[5].layers[1].norm1.bias",
            "[0].features.layers[5].layers[1].attn.relative_position_bias_table",
            "[1][<flat index 5>]",
            "[0].features.layers[5].layers[1].attn.qkv.weight",
            "[0].features.layers[5].layers[1].attn.qkv.bias",
            "[0].features.layers[5].layers[1].attn.proj.weight",
            "[0].features.layers[5].layers[1].attn.proj.bias",
            "[0].features.layers[5].layers[1].norm2.weight",
            "[0].features.layers[5].layers[1].norm2.bias",
            "[0].features.layers[5].layers[1].mlp.layers[0].weight",
            "[0].features.layers[5].layers[1].mlp.layers[0].bias",
            "[0].features.layers[5].layers[1].mlp.layers[3].weight",
            "[0].features.layers[5].layers[1].mlp.layers[3].bias",
            # Stage 3, Block 2 (features.5.2 -> layers[5].layers[2])
            "[0].features.layers[5].layers[2].norm1.weight",
            "[0].features.layers[5].layers[2].norm1.bias",
            "[0].features.layers[5].layers[2].attn.relative_position_bias_table",
            "[1][<flat index 6>]",
            "[0].features.layers[5].layers[2].attn.qkv.weight",
            "[0].features.layers[5].layers[2].attn.qkv.bias",
            "[0].features.layers[5].layers[2].attn.proj.weight",
            "[0].features.layers[5].layers[2].attn.proj.bias",
            "[0].features.layers[5].layers[2].norm2.weight",
            "[0].features.layers[5].layers[2].norm2.bias",
            "[0].features.layers[5].layers[2].mlp.layers[0].weight",
            "[0].features.layers[5].layers[2].mlp.layers[0].bias",
            "[0].features.layers[5].layers[2].mlp.layers[3].weight",
            "[0].features.layers[5].layers[2].mlp.layers[3].bias",
            # Stage 3, Block 3 (features.5.3 -> layers[5].layers[3])
            "[0].features.layers[5].layers[3].norm1.weight",
            "[0].features.layers[5].layers[3].norm1.bias",
            "[0].features.layers[5].layers[3].attn.relative_position_bias_table",
            "[1][<flat index 7>]",
            "[0].features.layers[5].layers[3].attn.qkv.weight",
            "[0].features.layers[5].layers[3].attn.qkv.bias",
            "[0].features.layers[5].layers[3].attn.proj.weight",
            "[0].features.layers[5].layers[3].attn.proj.bias",
            "[0].features.layers[5].layers[3].norm2.weight",
            "[0].features.layers[5].layers[3].norm2.bias",
            "[0].features.layers[5].layers[3].mlp.layers[0].weight",
            "[0].features.layers[5].layers[3].mlp.layers[0].bias",
            "[0].features.layers[5].layers[3].mlp.layers[3].weight",
            "[0].features.layers[5].layers[3].mlp.layers[3].bias",
            # Stage 3, Block 4 (features.5.4 -> layers[5].layers[4])
            "[0].features.layers[5].layers[4].norm1.weight",
            "[0].features.layers[5].layers[4].norm1.bias",
            "[0].features.layers[5].layers[4].attn.relative_position_bias_table",
            "[1][<flat index 8>]",
            "[0].features.layers[5].layers[4].attn.qkv.weight",
            "[0].features.layers[5].layers[4].attn.qkv.bias",
            "[0].features.layers[5].layers[4].attn.proj.weight",
            "[0].features.layers[5].layers[4].attn.proj.bias",
            "[0].features.layers[5].layers[4].norm2.weight",
            "[0].features.layers[5].layers[4].norm2.bias",
            "[0].features.layers[5].layers[4].mlp.layers[0].weight",
            "[0].features.layers[5].layers[4].mlp.layers[0].bias",
            "[0].features.layers[5].layers[4].mlp.layers[3].weight",
            "[0].features.layers[5].layers[4].mlp.layers[3].bias",
            # Stage 3, Block 5 (features.5.5 -> layers[5].layers[5])
            "[0].features.layers[5].layers[5].norm1.weight",
            "[0].features.layers[5].layers[5].norm1.bias",
            "[0].features.layers[5].layers[5].attn.relative_position_bias_table",
            "[1][<flat index 9>]",
            "[0].features.layers[5].layers[5].attn.qkv.weight",
            "[0].features.layers[5].layers[5].attn.qkv.bias",
            "[0].features.layers[5].layers[5].attn.proj.weight",
            "[0].features.layers[5].layers[5].attn.proj.bias",
            "[0].features.layers[5].layers[5].norm2.weight",
            "[0].features.layers[5].layers[5].norm2.bias",
            "[0].features.layers[5].layers[5].mlp.layers[0].weight",
            "[0].features.layers[5].layers[5].mlp.layers[0].bias",
            "[0].features.layers[5].layers[5].mlp.layers[3].weight",
            "[0].features.layers[5].layers[5].mlp.layers[3].bias",
            # Downsample 3 (features.6 -> layers[6])
            "[0].features.layers[6].reduction.weight",
            "[0].features.layers[6].norm.weight",
            "[0].features.layers[6].norm.bias",
            # Stage 4, Block 0 (features.7.0 -> layers[7].layers[0])
            "[0].features.layers[7].layers[0].norm1.weight",
            "[0].features.layers[7].layers[0].norm1.bias",
            "[0].features.layers[7].layers[0].attn.relative_position_bias_table",
            "[1][<flat index 10>]",
            "[0].features.layers[7].layers[0].attn.qkv.weight",
            "[0].features.layers[7].layers[0].attn.qkv.bias",
            "[0].features.layers[7].layers[0].attn.proj.weight",
            "[0].features.layers[7].layers[0].attn.proj.bias",
            "[0].features.layers[7].layers[0].norm2.weight",
            "[0].features.layers[7].layers[0].norm2.bias",
            "[0].features.layers[7].layers[0].mlp.layers[0].weight",
            "[0].features.layers[7].layers[0].mlp.layers[0].bias",
            "[0].features.layers[7].layers[0].mlp.layers[3].weight",
            "[0].features.layers[7].layers[0].mlp.layers[3].bias",
            "[0].features.layers[7].layers[1].norm1.weight",
            "[0].features.layers[7].layers[1].norm1.bias",
            "[0].features.layers[7].layers[1].attn.relative_position_bias_table",
            "[1][<flat index 11>]",
            "[0].features.layers[7].layers[1].attn.qkv.weight",
            "[0].features.layers[7].layers[1].attn.qkv.bias",
            "[0].features.layers[7].layers[1].attn.proj.weight",
            "[0].features.layers[7].layers[1].attn.proj.bias",
            "[0].features.layers[7].layers[1].norm2.weight",
            "[0].features.layers[7].layers[1].norm2.bias",
            "[0].features.layers[7].layers[1].mlp.layers[0].weight",
            "[0].features.layers[7].layers[1].mlp.layers[0].bias",
            "[0].features.layers[7].layers[1].mlp.layers[3].weight",
            "[0].features.layers[7].layers[1].mlp.layers[3].bias",
            "[0].norm.weight",
            "[0].norm.bias",
            "[0].head.weight",
            "[0].head.bias",
        ]
    return order



import functools
import os
from pathlib import Path
from urllib.request import urlretrieve

import equinox as eqx
import jax
import jax.numpy as jnp
from beartype.typing import Any, Callable, Literal, Sequence
from equinox.nn import StatefulLayer
from jaxonlayers.layers import ConvNormActivation, LayerNorm, StochasticDepth
from jaxonlayers.layers.abstract import AbstractNorm
from jaxtyping import Array, Float, PRNGKeyArray
from statedict2pytree import (
    convert,
    move_running_fields_to_the_end,
    pytree_to_fields,
    state_dict_to_fields,
)

from jaxonmodels.functions import default_floating_dtype, dtype_to_str

_MODELS = {
    "convnext_tiny_IMAGENET1K_V1": "https://download.pytorch.org/models/convnext_tiny-983f1562.pth",
    "convnext_small_IMAGENET1K_V1": "https://download.pytorch.org/models/convnext_small-0c510722.pth",
    "convnext_base_IMAGENET1K_V1": "https://download.pytorch.org/models/convnext_base-6075fbad.pth",
    "convnext_large_IMAGENET1K_V1": "https://download.pytorch.org/models/convnext_large-ea097f82.pth",
}


class CNBlockConfig:
    # Stores information listed at Section 3 of the ConvNeXt paper
    def __init__(
        self,
        input_channels: int,
        out_channels: int | None,
        num_layers: int,
    ) -> None:
        self.input_channels = input_channels
        self.out_channels = out_channels
        self.num_layers = num_layers

    def __repr__(self) -> str:
        s = self.__class__.__name__ + "("
        s += "input_channels={input_channels}"
        s += ", out_channels={out_channels}"
        s += ", num_layers={num_layers}"
        s += ")"
        return s.format(**self.__dict__)


class CNBlock(eqx.Module):
    layer_scale: Array
    dwconv: eqx.nn.Conv2d
    norm: AbstractNorm
    pwconv1: eqx.nn.Linear
    pwconv2: eqx.nn.Linear
    stochastic_depth: StochasticDepth

    def __init__(
        self,
        dim: int,
        layer_scale: float,
        stochastic_depth_prob: float,
        norm_layer: Callable[..., AbstractNorm] | None,
        *,
        key: PRNGKeyArray,
        dtype: Any,
    ) -> None:
        key, *keys = jax.random.split(key, 5)

        self.layer_scale = jnp.ones((dim, 1, 1), dtype=dtype) * layer_scale
        self.dwconv = eqx.nn.Conv2d(
            in_channels=dim,
            out_channels=dim,
            kernel_size=7,
            padding=3,
            groups=dim,
            use_bias=True,
            key=keys[0],
            dtype=dtype,
        )
        if norm_layer is None:
            self.norm = LayerNorm(shape=dim, eps=1e-6)
        else:
            self.norm = norm_layer(shape=dim, eps=1e-6)

        assert self.norm is not None

        self.pwconv1 = eqx.nn.Linear(
            in_features=dim,
            out_features=4 * dim,
            use_bias=True,
            key=keys[2],
            dtype=dtype,
        )

        self.pwconv2 = eqx.nn.Linear(
            in_features=4 * dim,
            out_features=dim,
            use_bias=True,
            key=keys[3],
            dtype=dtype,
        )
        self.stochastic_depth = StochasticDepth(p=stochastic_depth_prob, mode="row")

    def __call__(
        self,
        x: Float[Array, "c h w"],
        key: PRNGKeyArray,
    ) -> Float[Array, "c h w"]:
        residual = x
        x = self.dwconv(x)
        x = jnp.transpose(x, (1, 2, 0))
        x = self.norm(x)  # pyright: ignore
        x = eqx.filter_vmap(eqx.filter_vmap(self.pwconv1))(x)
        x = jax.nn.gelu(x)
        x = eqx.filter_vmap(eqx.filter_vmap(self.pwconv2))(x)
        x = jnp.transpose(x, (2, 0, 1))
        x = self.layer_scale * x
        x = self.stochastic_depth(x, key=key)
        x = x + residual
        return x


class ConvNeXt(StatefulLayer):
    features: eqx.nn.Sequential
    avgpool: eqx.nn.AdaptiveAvgPool2d
    classifier: eqx.nn.Sequential

    inference: bool

    def __init__(
        self,
        block_setting: list[CNBlockConfig],
        stochastic_depth_prob: float,
        layer_scale: float,
        num_classes: int,
        block: Callable[..., eqx.Module] | None,
        norm_layer: Callable[..., AbstractNorm] | None,
        inference: bool,
        key: PRNGKeyArray,
        dtype: Any,
    ) -> None:
        self.inference = inference
        if not block_setting:
            raise ValueError("The block_setting should not be empty")
        elif not (
            isinstance(block_setting, Sequence)
            and all([isinstance(s, CNBlockConfig) for s in block_setting])
        ):
            raise TypeError("The block_setting should be List[CNBlockConfig]")

        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None

        if block is None:
            block = CNBlock

        key, subkey = jax.random.split(key)
        firstconv_output_channels = block_setting[0].input_channels

        if norm_layer is None:
            norm_layer = functools.partial(LayerNorm, dtype=dtype)
        assert norm_layer is not None
        layers: list[eqx.Module] = []

        layers.append(
            ConvNormActivation(
                2,
                3,
                firstconv_output_channels,
                kernel_size=4,
                stride=4,
                padding=0,
                norm_layer=functools.partial(
                    norm_layer, shape=firstconv_output_channels
                ),
                activation_layer=None,
                use_bias=True,
                key=subkey,
                dtype=dtype,
            )
        )

        total_stage_blocks = sum(cnf.num_layers for cnf in block_setting)
        stage_block_id = 0
        for cnf in block_setting:
            # Bottlenecks
            stage: list[eqx.Module] = []
            key, *subkeys = jax.random.split(key, cnf.num_layers + 5)
            for i in range(cnf.num_layers):
                sd_prob = (
                    stochastic_depth_prob * stage_block_id / (total_stage_blocks - 1.0)
                )
                stage.append(
                    block(
                        cnf.input_channels,
                        layer_scale,
                        sd_prob,
                        norm_layer=None,
                        key=subkeys[i],
                        dtype=dtype,
                    )
                )
                stage_block_id += 1
            layers.append(eqx.nn.Sequential(stage))  # pyright: ignore
            if cnf.out_channels is not None:
                # Downsampling
                key, subkey = jax.random.split(key)
                layers.append(
                    eqx.nn.Sequential(
                        [
                            norm_layer(cnf.input_channels),  # pyright: ignore
                            eqx.nn.Conv2d(
                                cnf.input_channels,
                                cnf.out_channels,
                                kernel_size=2,
                                stride=2,
                                dtype=dtype,
                                key=subkey,
                            ),
                        ]
                    )
                )

        self.features = eqx.nn.Sequential(layers)  # pyright: ignore
        self.avgpool = eqx.nn.AdaptiveAvgPool2d(1)

        lastblock = block_setting[-1]
        lastconv_output_channels = (
            lastblock.out_channels
            if lastblock.out_channels is not None
            else lastblock.input_channels
        )
        key, subkey = jax.random.split(key)
        self.classifier = eqx.nn.Sequential(
            [
                norm_layer(lastconv_output_channels),  # pyright: ignore
                eqx.nn.Lambda(lambda x: jnp.ravel(x)),
                eqx.nn.Linear(
                    lastconv_output_channels, num_classes, key=subkey, dtype=dtype
                ),
            ]
        )

        # todo: init properly

    def __call__(
        self,
        x: Array,
        state: eqx.nn.State,
        key: PRNGKeyArray,
    ):
        key, *subkeys = jax.random.split(key, 4)
        x, state = self.features(x, key=subkeys[0], state=state)
        x = self.avgpool(x, key=subkeys[1])
        x, state = self.classifier(x, key=subkeys[2], state=state)
        return x, state


def _convnext_tiny(key, inference: bool, num_classes=1000, dtype: Any = None):
    """Creates a ConvNeXt Tiny model."""
    if dtype is None:
        dtype = default_floating_dtype()
    assert dtype is not None

    block_setting = [
        CNBlockConfig(96, 192, 3),
        CNBlockConfig(192, 384, 3),
        CNBlockConfig(384, 768, 9),
        CNBlockConfig(768, None, 3),
    ]

    model, state = eqx.nn.make_with_state(ConvNeXt)(
        block_setting=block_setting,
        stochastic_depth_prob=0.1,
        num_classes=num_classes,
        layer_scale=1e-6,
        block=None,
        norm_layer=None,
        inference=inference,
        key=key,
        dtype=dtype,
    )

    return model, state


def _convnext_small(key, inference: bool, num_classes=1000, dtype: Any = None):
    """Creates a ConvNeXt Small model."""
    if dtype is None:
        dtype = default_floating_dtype()
    assert dtype is not None

    block_setting = [
        CNBlockConfig(96, 192, 3),
        CNBlockConfig(192, 384, 3),
        CNBlockConfig(384, 768, 27),
        CNBlockConfig(768, None, 3),
    ]

    model, state = eqx.nn.make_with_state(ConvNeXt)(
        block_setting=block_setting,
        stochastic_depth_prob=0.4,
        num_classes=num_classes,
        layer_scale=1e-6,
        block=None,
        norm_layer=None,
        inference=inference,
        key=key,
        dtype=dtype,
    )

    return model, state


def _convnext_base(key, inference: bool, num_classes=1000, dtype: Any = None):
    """Creates a ConvNeXt Base model."""
    if dtype is None:
        dtype = default_floating_dtype()
    assert dtype is not None

    block_setting = [
        CNBlockConfig(128, 256, 3),
        CNBlockConfig(256, 512, 3),
        CNBlockConfig(512, 1024, 27),
        CNBlockConfig(1024, None, 3),
    ]

    model, state = eqx.nn.make_with_state(ConvNeXt)(
        block_setting=block_setting,
        stochastic_depth_prob=0.5,
        num_classes=num_classes,
        layer_scale=1e-6,
        block=None,
        norm_layer=None,
        inference=inference,
        key=key,
        dtype=dtype,
    )

    return model, state


def _convnext_large(key, inference: bool, num_classes=1000, dtype: Any = None):
    """Creates a ConvNeXt Large model."""
    if dtype is None:
        dtype = default_floating_dtype()
    assert dtype is not None

    block_setting = [
        CNBlockConfig(192, 384, 3),
        CNBlockConfig(384, 768, 3),
        CNBlockConfig(768, 1536, 27),
        CNBlockConfig(1536, None, 3),
    ]

    model, state = eqx.nn.make_with_state(ConvNeXt)(
        block_setting=block_setting,
        stochastic_depth_prob=0.5,
        num_classes=num_classes,
        layer_scale=1e-6,
        block=None,
        norm_layer=None,
        inference=inference,
        key=key,
        dtype=dtype,
    )

    return model, state


def _with_weights(
    model,
    weights: str,
    cache: bool,
    dtype: Any,
):
    """Load pre-trained weights for the model."""
    dtype_str = dtype_to_str(dtype)

    weights_url = _MODELS.get(weights)
    if weights_url is None:
        raise ValueError(f"No weights found for {weights}")

    # Create directories for caching
    jaxonmodels_dir = os.path.expanduser("~/.jaxonmodels/models")
    os.makedirs(jaxonmodels_dir, exist_ok=True)

    # Define cache file path including dtype
    cache_filename = f"{weights}-{dtype_str}.eqx"
    cache_filepath = str(Path(jaxonmodels_dir) / cache_filename)

    # Check if cached model exists
    if cache:
        if os.path.exists(cache_filepath):
            return eqx.tree_deserialise_leaves(
                cache_filepath,
                model,
            )

    # Download weights if not already downloaded
    weights_dir = os.path.expanduser("~/.jaxonmodels/pytorch_weights")
    os.makedirs(weights_dir, exist_ok=True)
    filename = weights_url.split("/")[-1]
    weights_file = os.path.join(weights_dir, filename)
    if not os.path.exists(weights_file):
        print(f"Downloading weights from {weights_url} to {weights_file}")
        urlretrieve(weights_url, weights_file)
    else:
        print(f"Using existing weights file: {weights_file}")

    # Load PyTorch weights
    import torch

    print("Loading PyTorch state dict...")
    weights_dict = torch.load(
        weights_file,
        map_location=torch.device("cpu"),  # weights_only=True removed for compatibility
    )
    # If it's a checkpoint dict, extract the state_dict
    if not isinstance(weights_dict, dict):
        # Try loading as a JIT model like in CLIP if simple load fails
        try:
            torch_model = torch.jit.load(weights_file, map_location=torch.device("cpu"))
            temp_dict = {}
            for name, param in torch_model.named_parameters():
                temp_dict[name] = param.clone().detach()
            for name, buffer in torch_model.named_buffers():
                temp_dict[name] = buffer.clone().detach()
            weights_dict = temp_dict
            print("Loaded weights from TorchScript model.")
        except Exception as e:
            raise TypeError(
                f"Could not load weights. "
                f"Expected state_dict but got {type(weights_dict)}. "
                f"JIT load failed with: {e}"
            )
    elif "state_dict" in weights_dict:
        weights_dict = weights_dict["state_dict"]

    print("Converting PyTorch weights to JAX format...")
    # Convert PyTorch weights to JAX format
    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    jaxfields, state_indices = pytree_to_fields(model)

    model = convert(
        weights_dict,
        model,
        jaxfields,
        state_indices,
        torchfields,
        dtype=dtype,
    )

    if cache:
        eqx.tree_serialise_leaves(cache_filepath, model)

    return model


def _assert_model_and_weights_fit(
    model: Literal[
        "convnext_tiny",
        "convnext_small",
        "convnext_base",
        "convnext_large",
    ],
    weights: Literal[
        "convnext_tiny_IMAGENET1K_V1",
        "convnext_small_IMAGENET1K_V1",
        "convnext_base_IMAGENET1K_V1",
        "convnext_large_IMAGENET1K_V1",
    ],
):
    """Ensures the model and weights are compatible."""
    if "_" in weights:
        weights_model = weights.split("_IMAGENET")[0].lower()

        # Check if the specified weights are compatible with the model
        if weights_model != model:
            raise ValueError(
                f"Model {model} is incompatible with weights {weights}. "
                f"Expected weights starting with '{model}'."
            )


def load_convnext(
    model: Literal[
        "convnext_tiny",
        "convnext_small",
        "convnext_base",
        "convnext_large",
    ],
    weights: Literal[
        "convnext_tiny_IMAGENET1K_V1",
        "convnext_small_IMAGENET1K_V1",
        "convnext_base_IMAGENET1K_V1",
        "convnext_large_IMAGENET1K_V1",
    ]
    | None = None,
    num_classes: int = 1000,
    cache: bool = True,
    inference: bool = False,
    key: PRNGKeyArray | None = None,
    dtype: Any = None,
) -> tuple[ConvNeXt, eqx.nn.State]:
    """
    Load a ConvNeXt model with optional pre-trained weights.

    Args:
        model: The name of the ConvNeXt model to load
        weights: The name of the weights to load (from _MODELS), or None for random init
        num_classes: Number of output classes
        cache: Whether to cache the model and weights.
        key: Random key for initialization. Defaults to jax.random.key(42).
        dtype: The data type for the model's parameters (e.g., jnp.float32).
               Defaults to jaxonmodels.functions.default_floating_dtype().

    Returns:
        A ConvNeXt model
    """
    if key is None:
        key = jax.random.key(42)

    # Determine default dtype if not provided before passing to constructors
    if dtype is None:
        dtype = default_floating_dtype()
    assert dtype is not None  # Ensure dtype is set

    convnext = None
    state = None

    match model:
        case "convnext_tiny":
            convnext, state = _convnext_tiny(key, inference, num_classes, dtype=dtype)
        case "convnext_small":
            convnext, state = _convnext_small(key, inference, num_classes, dtype=dtype)
        case "convnext_base":
            convnext, state = _convnext_base(key, inference, num_classes, dtype=dtype)
        case "convnext_large":
            convnext, state = _convnext_large(key, inference, num_classes, dtype=dtype)
        case _:
            raise ValueError(f"Unknown model name: {model}")

    if weights:
        _assert_model_and_weights_fit(model, weights)
        convnext, state = _with_weights((convnext, state), weights, cache, dtype=dtype)

    assert convnext is not None

    return convnext, state



import copy
import functools
import math
import os
from dataclasses import dataclass
from pathlib import Path
from urllib.request import urlretrieve

import equinox as eqx
import jax
import jax.numpy as jnp
from beartype.typing import Any, Callable, Literal, Sequence
from jaxonlayers.layers import (
    BatchNorm,
    ConvNormActivation,
    SqueezeExcitation,
    StochasticDepth,
)
from jaxtyping import Array, PRNGKeyArray
from statedict2pytree import (
    convert,
    move_running_fields_to_the_end,
    pytree_to_fields,
    state_dict_to_fields,
)

from jaxonmodels.functions import (
    default_floating_dtype,
    make_divisible,
)
from jaxonmodels.functions.utils import dtype_to_str

_MODELS = {
    "efficientnet_b0_IMAGENET1K_V1": "https://download.pytorch.org/models/efficientnet_b0_rwightman-7f5810bc.pth",
    "efficientnet_b1_IMAGENET1K_V1": "https://download.pytorch.org/models/efficientnet_b1_rwightman-bac287d4.pth",
    "efficientnet_b1_IMAGENET1K_V2": "https://download.pytorch.org/models/efficientnet_b1-c27df63c.pth",
    "efficientnet_b2_IMAGENET1K_V1": "https://download.pytorch.org/models/efficientnet_b2_rwightman-c35c1473.pth",
    "efficientnet_b3_IMAGENET1K_V1": "https://download.pytorch.org/models/efficientnet_b3_rwightman-b3899882.pth",
    "efficientnet_b4_IMAGENET1K_V1": "https://download.pytorch.org/models/efficientnet_b4_rwightman-23ab8bcd.pth",
    "efficientnet_b5_IMAGENET1K_V1": "https://download.pytorch.org/models/efficientnet_b5_lukemelas-1a07897c.pth",
    "efficientnet_b6_IMAGENET1K_V1": "https://download.pytorch.org/models/efficientnet_b6_lukemelas-24a108a5.pth",
    "efficientnet_b7_IMAGENET1K_V1": "https://download.pytorch.org/models/efficientnet_b7_lukemelas-c5b4e57e.pth",
    "efficientnet_v2_s_IMAGENET1K_V1": "https://download.pytorch.org/models/efficientnet_v2_s-dd5fe13b.pth",
    "efficientnet_v2_m_IMAGENET1K_V1": "https://download.pytorch.org/models/efficientnet_v2_m-dc08266a.pth",
    "efficientnet_v2_l_IMAGENET1K_V1": "https://download.pytorch.org/models/efficientnet_v2_l-59c71312.pth",
}


@dataclass
class _MBConvConfig:
    expand_ratio: float
    kernel: int
    stride: int
    input_channels: int
    out_channels: int
    num_layers: int
    block: Callable[..., "MBConv | FusedMBConv"]

    @staticmethod
    def adjust_channels(
        channels: int, width_mult: float, min_value: int | None = None
    ) -> int:
        return make_divisible(channels * width_mult, 8, min_value)


class MBConvConfig(_MBConvConfig):
    # Stores information listed at Table 1 of the EfficientNet paper &
    # Table 4 of the EfficientNetV2 paper
    def __init__(
        self,
        expand_ratio: float,
        kernel: int,
        stride: int,
        input_channels: int,
        out_channels: int,
        num_layers: int,
        width_mult: float = 1.0,
        depth_mult: float = 1.0,
    ) -> None:
        input_channels = self.adjust_channels(input_channels, width_mult)
        out_channels = self.adjust_channels(out_channels, width_mult)
        num_layers = self.adjust_depth(num_layers, depth_mult)
        super().__init__(
            expand_ratio,
            kernel,
            stride,
            input_channels,
            out_channels,
            num_layers,
            MBConv,
        )

    @staticmethod
    def adjust_depth(num_layers: int, depth_mult: float):
        return int(math.ceil(num_layers * depth_mult))


class FusedMBConvConfig(_MBConvConfig):
    # Stores information listed at Table 4 of the EfficientNetV2 paper
    def __init__(
        self,
        expand_ratio: float,
        kernel: int,
        stride: int,
        input_channels: int,
        out_channels: int,
        num_layers: int,
    ) -> None:
        super().__init__(
            expand_ratio,
            kernel,
            stride,
            input_channels,
            out_channels,
            num_layers,
            FusedMBConv,
        )


class MBConv(eqx.Module):
    use_res_connect: bool = eqx.field(static=True)

    expand_conv2d: ConvNormActivation | None
    depthwise_conv2d: ConvNormActivation
    se_layer: SqueezeExcitation

    project_conv2d: ConvNormActivation
    stochastic_depth: StochasticDepth

    def __init__(
        self,
        cnf: MBConvConfig,
        stochastic_depth_prob: float,
        key: PRNGKeyArray,
        dtype: Any,
        axis_name: str = "batch",
    ):
        if not (1 <= cnf.stride <= 2):
            raise ValueError("illegal stride value")

        self.use_res_connect = (
            cnf.stride == 1 and cnf.input_channels == cnf.out_channels
        )

        activation_layer = jax.nn.silu

        # expand
        expanded_channels = cnf.adjust_channels(cnf.input_channels, cnf.expand_ratio)
        self.expand_conv2d = None
        if expanded_channels != cnf.input_channels:
            key, subkey = jax.random.split(key)
            self.expand_conv2d = ConvNormActivation(
                2,
                cnf.input_channels,
                expanded_channels,
                kernel_size=1,
                norm_layer=functools.partial(
                    BatchNorm, size=expanded_channels, axis_name=axis_name, dtype=dtype
                ),
                activation_layer=activation_layer,
                key=subkey,
                dtype=dtype,
            )

        # depthwise
        key, subkey = jax.random.split(key)
        self.depthwise_conv2d = ConvNormActivation(
            2,
            expanded_channels,
            expanded_channels,
            kernel_size=cnf.kernel,
            stride=cnf.stride,
            groups=expanded_channels,
            norm_layer=functools.partial(
                BatchNorm, size=expanded_channels, axis_name=axis_name, dtype=dtype
            ),
            activation_layer=activation_layer,
            key=subkey,
            dtype=dtype,
        )
        squeeze_channels = max(1, cnf.input_channels // 4)

        key, subkey = jax.random.split(key)
        self.se_layer = SqueezeExcitation(
            expanded_channels, squeeze_channels, key=subkey, dtype=dtype
        )

        # project
        key, subkey = jax.random.split(key)
        self.project_conv2d = ConvNormActivation(
            2,
            expanded_channels,
            cnf.out_channels,
            kernel_size=1,
            norm_layer=functools.partial(
                BatchNorm, size=cnf.out_channels, axis_name=axis_name, dtype=dtype
            ),
            activation_layer=None,
            key=subkey,
            dtype=dtype,
        )
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, "row")

    def __call__(
        self, x: Array, state: eqx.nn.State, key: PRNGKeyArray
    ) -> tuple[Array, eqx.nn.State]:
        if self.expand_conv2d:
            result, state = self.expand_conv2d(x, state)
        else:
            result = x
        result, state = self.depthwise_conv2d(result, state)
        result = self.se_layer(result, activation=jax.nn.silu)

        result, state = self.project_conv2d(result, state)

        if self.use_res_connect:
            result = self.stochastic_depth(result, key=key)
            result += x
        return result, state


class FusedMBConv(eqx.Module):
    fused_expand: ConvNormActivation
    project: ConvNormActivation | None
    stochastic_depth: StochasticDepth

    use_res_connect: bool = eqx.field(static=True)

    def __init__(
        self,
        cnf: FusedMBConvConfig,
        stochastic_depth_prob: float,
        key: PRNGKeyArray,
        dtype: Any,
        axis_name: str = "batch",
    ):
        if not (1 <= cnf.stride <= 2):
            raise ValueError("illegal stride value")

        self.use_res_connect = (
            cnf.stride == 1 and cnf.input_channels == cnf.out_channels
        )

        expanded_channels = cnf.adjust_channels(cnf.input_channels, cnf.expand_ratio)
        self.project = None
        if expanded_channels != cnf.input_channels:
            # fused expand
            key, subkey, subkey2 = jax.random.split(key, 3)
            self.fused_expand = ConvNormActivation(
                2,
                cnf.input_channels,
                expanded_channels,
                kernel_size=cnf.kernel,
                stride=cnf.stride,
                norm_layer=functools.partial(
                    BatchNorm, size=expanded_channels, axis_name=axis_name, dtype=dtype
                ),
                activation_layer=jax.nn.silu,
                key=subkey,
                dtype=dtype,
            )

            # project
            self.project = ConvNormActivation(
                2,
                expanded_channels,
                cnf.out_channels,
                kernel_size=1,
                norm_layer=functools.partial(
                    BatchNorm, size=cnf.out_channels, axis_name=axis_name, dtype=dtype
                ),
                activation_layer=None,
                key=subkey2,
                dtype=dtype,
            )
        else:
            key, subkey = jax.random.split(key)
            self.fused_expand = ConvNormActivation(
                2,
                cnf.input_channels,
                cnf.out_channels,
                kernel_size=cnf.kernel,
                stride=cnf.stride,
                norm_layer=functools.partial(
                    BatchNorm, size=cnf.out_channels, axis_name=axis_name, dtype=dtype
                ),
                activation_layer=jax.nn.silu,
                key=subkey,
                dtype=dtype,
            )
            self.project = None

        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, "row")

    def __call__(
        self, x: Array, state: eqx.nn.State, *, key: PRNGKeyArray
    ) -> tuple[Array, eqx.nn.State]:
        result, state = self.fused_expand(x, state)

        if self.project:
            result, state = self.project(result, state)

        if self.use_res_connect:
            result = self.stochastic_depth(result, key=key)
            result += x
        return result, state


class InvertedResidualBlock(eqx.Module):
    layers: list[MBConv | FusedMBConv]

    def __init__(self, layers: list[MBConv | FusedMBConv]):
        self.layers = layers

    def __call__(
        self, x: Array, state: eqx.nn.State, key: PRNGKeyArray
    ) -> tuple[Array, eqx.nn.State]:
        keys = jax.random.split(key, len(self.layers) + 1)
        for i, stage in enumerate(self.layers):
            x, state = stage(x, state, key=keys[i])

        return x, state


class Classifier(eqx.Module):
    dropout: eqx.nn.Dropout
    linear: eqx.nn.Linear

    def __init__(
        self,
        p: float,
        in_features: int,
        out_features: int,
        key: PRNGKeyArray,
        dtype: Any,
    ):
        self.dropout = eqx.nn.Dropout(p=p)

        self.linear = eqx.nn.Linear(in_features, out_features, key=key, dtype=dtype)

    def __call__(self, x: Array, key: PRNGKeyArray) -> Array:
        x = self.dropout(x, key=key)
        x = self.linear(x)
        return x


class EfficientNet(eqx.Module):
    features: list[ConvNormActivation | MBConv | FusedMBConv | InvertedResidualBlock]
    avgpool: eqx.nn.AdaptiveAvgPool2d
    classifier: Classifier

    def __init__(
        self,
        inverted_residual_setting: Sequence[MBConvConfig | FusedMBConvConfig],
        dropout: float,
        stochastic_depth_prob: float = 0.2,
        n_classes: int = 1000,
        last_channel: int | None = None,
        *,
        key: PRNGKeyArray,
        axis_name: str = "batch",
        dtype: Any | None = None,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None

        if not inverted_residual_setting:
            raise ValueError("The inverted_residual_setting should not be empty")
        elif not (
            isinstance(inverted_residual_setting, Sequence)
            and all([isinstance(s, _MBConvConfig) for s in inverted_residual_setting])
        ):
            raise TypeError(
                "The inverted_residual_setting should be List[MBConvConfig]"
                " or List[FusedMBConvConfig]"
            )

        key, *subkeys = jax.random.split(key, 10)

        firstconv_output_channels = inverted_residual_setting[0].input_channels
        self.features = []
        self.features.append(
            ConvNormActivation(
                2,
                3,
                firstconv_output_channels,
                kernel_size=3,
                stride=2,
                norm_layer=functools.partial(
                    BatchNorm,
                    size=firstconv_output_channels,
                    axis_name=axis_name,
                    dtype=dtype,
                ),
                activation_layer=jax.nn.silu,
                key=subkeys[0],
                dtype=dtype,
            )
        )

        # building inverted residual blocks
        total_stage_blocks = sum(cnf.num_layers for cnf in inverted_residual_setting)
        stage_block_id = 0
        for cnf in inverted_residual_setting:
            stage: list[MBConv | FusedMBConv] = []
            num_layers = cnf.num_layers
            stage_keys = jax.random.split(key, num_layers + 1)
            key = stage_keys[0]
            for i in range(num_layers):
                block_cnf = copy.copy(cnf)

                if stage:
                    block_cnf.input_channels = block_cnf.out_channels
                    block_cnf.stride = 1

                sd_prob = (
                    stochastic_depth_prob * float(stage_block_id) / total_stage_blocks
                )
                stage.append(
                    block_cnf.block(block_cnf, sd_prob, stage_keys[i], dtype=dtype)
                )
                stage_block_id += 1

            self.features.append(InvertedResidualBlock(stage))

        # building last several layers
        lastconv_input_channels = inverted_residual_setting[-1].out_channels
        lastconv_output_channels = (
            last_channel if last_channel is not None else 4 * lastconv_input_channels
        )
        key, subkey1, subkey2 = jax.random.split(key, 3)
        self.features.append(
            ConvNormActivation(
                2,
                lastconv_input_channels,
                lastconv_output_channels,
                kernel_size=1,
                norm_layer=functools.partial(
                    BatchNorm,
                    size=lastconv_output_channels,
                    axis_name=axis_name,
                    dtype=dtype,
                ),
                activation_layer=jax.nn.silu,
                key=subkey1,
                dtype=dtype,
            )
        )

        self.avgpool = eqx.nn.AdaptiveAvgPool2d(1)
        self.classifier = Classifier(
            dropout,
            lastconv_output_channels,
            n_classes,
            subkey2,
            dtype=dtype,
        )
        # TODO: kaiming init

    def __call__(
        self, x: Array, state: eqx.nn.State, key: PRNGKeyArray
    ) -> tuple[Array, eqx.nn.State]:
        num_feature_layers = len(self.features)
        keys = jax.random.split(key, num_feature_layers + 2)
        feature_keys = keys[:num_feature_layers]
        classifier_key = keys[-1]

        for i, feature in enumerate(self.features):
            x, state = feature(x, state, key=feature_keys[i])

        x = self.avgpool(x)
        x = jnp.ravel(x)

        x = self.classifier(x, classifier_key)

        return x, state


def _efficientnet(
    inverted_residual_setting: Sequence[MBConvConfig | FusedMBConvConfig],
    dropout: float,
    last_channel: int | None,
    n_classes: int,
    key: PRNGKeyArray,
    dtype: Any,
) -> tuple[EfficientNet, eqx.nn.State]:
    model, state = eqx.nn.make_with_state(EfficientNet)(
        inverted_residual_setting,
        dropout,
        last_channel=last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )

    return model, state


def _efficientnet_conf(
    arch: str,
    **kwargs: Any,
) -> tuple[Sequence[MBConvConfig | FusedMBConvConfig], int | None]:
    inverted_residual_setting: Sequence[MBConvConfig | FusedMBConvConfig]
    if arch.startswith("efficientnet_b"):
        bneck_conf = functools.partial(
            MBConvConfig,
            width_mult=kwargs.pop("width_mult"),
            depth_mult=kwargs.pop("depth_mult"),
        )
        inverted_residual_setting = [
            bneck_conf(1, 3, 1, 32, 16, 1),
            bneck_conf(6, 3, 2, 16, 24, 2),
            bneck_conf(6, 5, 2, 24, 40, 2),
            bneck_conf(6, 3, 2, 40, 80, 3),
            bneck_conf(6, 5, 1, 80, 112, 3),
            bneck_conf(6, 5, 2, 112, 192, 4),
            bneck_conf(6, 3, 1, 192, 320, 1),
        ]
        last_channel = None
    elif arch.startswith("efficientnet_v2_s"):
        inverted_residual_setting = [
            FusedMBConvConfig(1, 3, 1, 24, 24, 2),
            FusedMBConvConfig(4, 3, 2, 24, 48, 4),
            FusedMBConvConfig(4, 3, 2, 48, 64, 4),
            MBConvConfig(4, 3, 2, 64, 128, 6),
            MBConvConfig(6, 3, 1, 128, 160, 9),
            MBConvConfig(6, 3, 2, 160, 256, 15),
        ]
        last_channel = 1280
    elif arch.startswith("efficientnet_v2_m"):
        inverted_residual_setting = [
            FusedMBConvConfig(1, 3, 1, 24, 24, 3),
            FusedMBConvConfig(4, 3, 2, 24, 48, 5),
            FusedMBConvConfig(4, 3, 2, 48, 80, 5),
            MBConvConfig(4, 3, 2, 80, 160, 7),
            MBConvConfig(6, 3, 1, 160, 176, 14),
            MBConvConfig(6, 3, 2, 176, 304, 18),
            MBConvConfig(6, 3, 1, 304, 512, 5),
        ]
        last_channel = 1280
    elif arch.startswith("efficientnet_v2_l"):
        inverted_residual_setting = [
            FusedMBConvConfig(1, 3, 1, 32, 32, 4),
            FusedMBConvConfig(4, 3, 2, 32, 64, 7),
            FusedMBConvConfig(4, 3, 2, 64, 96, 7),
            MBConvConfig(4, 3, 2, 96, 192, 10),
            MBConvConfig(6, 3, 1, 192, 224, 19),
            MBConvConfig(6, 3, 2, 224, 384, 25),
            MBConvConfig(6, 3, 1, 384, 640, 7),
        ]
        last_channel = 1280
    else:
        raise ValueError(f"Unsupported model type {arch}")

    return inverted_residual_setting, last_channel


def _with_weights(
    pytree,
    weights: str,
    cache: bool,
    dtype: Any,
):
    """Load pre-trained weights for the model"""
    efficientnet, state = pytree
    dtype_str = dtype_to_str(dtype)

    weights_url = _MODELS.get(weights)
    if weights_url is None:
        raise ValueError(f"No weights found for {weights}")

    # Create directories for caching
    jaxonmodels_dir = os.path.expanduser("~/.jaxonmodels/models")
    os.makedirs(jaxonmodels_dir, exist_ok=True)

    # Define cache file path including dtype
    cache_filename = f"{weights}-{dtype_str}.eqx"
    cache_filepath = str(Path(jaxonmodels_dir) / cache_filename)

    # Check if cached model exists
    if cache:
        if os.path.exists(cache_filepath):
            return eqx.tree_deserialise_leaves(
                cache_filepath,
                (efficientnet, state),
            )

    # Download weights if not already downloaded
    weights_dir = os.path.expanduser("~/.jaxonmodels/pytorch_weights")
    os.makedirs(weights_dir, exist_ok=True)
    filename = weights_url.split("/")[-1]
    weights_file = os.path.join(weights_dir, filename)
    if not os.path.exists(weights_file):
        print(f"Downloading weights from {weights_url} to {weights_file}")
        urlretrieve(weights_url, weights_file)
    else:
        print(f"Using existing weights file: {weights_file}")

    # Load PyTorch weights
    import torch

    print("Loading PyTorch state dict...")
    weights_dict = torch.load(
        weights_file,
        map_location=torch.device("cpu"),  # weights_only=True removed for compatibility
    )
    # If it's a checkpoint dict, extract the state_dict
    if not isinstance(weights_dict, dict):
        try:
            torch_model = torch.jit.load(weights_file, map_location=torch.device("cpu"))
            temp_dict = {}
            for name, param in torch_model.named_parameters():
                temp_dict[name] = param.clone().detach()
            for name, buffer in torch_model.named_buffers():
                temp_dict[name] = buffer.clone().detach()
            # EfficientNet doesn't have logit_scale like CLIP
            weights_dict = temp_dict
            print("Loaded weights from TorchScript model.")
        except Exception as e:
            raise TypeError(
                f"Could not load weights. "
                f"Expected state_dict but got {type(weights_dict)}. "
                f"JIT load failed with: {e}"
            )
    elif "state_dict" in weights_dict:
        weights_dict = weights_dict["state_dict"]

    print("Converting PyTorch weights to JAX format...")
    # Convert PyTorch weights to JAX format
    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    jaxfields, state_indices = pytree_to_fields((efficientnet, state))

    efficientnet, state = convert(
        weights_dict,
        (efficientnet, state),
        jaxfields,
        state_indices,
        torchfields,
        dtype=dtype,
    )

    if cache:
        eqx.tree_serialise_leaves(cache_filepath, (efficientnet, state))

    return efficientnet, state


def _efficientnet_b0(key, n_classes=1000, dtype: Any = None):
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b0", width_mult=1.0, depth_mult=1.0
    )
    return _efficientnet(
        inverted_residual_setting,
        0.2,
        last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )


def _efficientnet_b1(key, n_classes=1000, dtype: Any = None):
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b1", width_mult=1.0, depth_mult=1.1
    )
    return _efficientnet(
        inverted_residual_setting,
        0.2,
        last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )


def _efficientnet_b2(key, n_classes=1000, dtype: Any = None):
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b2", width_mult=1.1, depth_mult=1.2
    )
    return _efficientnet(
        inverted_residual_setting,
        0.3,
        last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )


def _efficientnet_b3(key, n_classes=1000, dtype: Any = None):
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b3", width_mult=1.2, depth_mult=1.4
    )
    return _efficientnet(
        inverted_residual_setting,
        0.3,
        last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )


def _efficientnet_b4(key, n_classes=1000, dtype: Any = None):
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b4", width_mult=1.4, depth_mult=1.8
    )
    return _efficientnet(
        inverted_residual_setting,
        0.4,
        last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )


def _efficientnet_b5(key, n_classes=1000, dtype: Any = None):
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b5", width_mult=1.6, depth_mult=2.2
    )
    return _efficientnet(
        inverted_residual_setting,
        0.4,
        last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )


def _efficientnet_b6(key, n_classes=1000, dtype: Any = None):
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b6", width_mult=1.8, depth_mult=2.6
    )
    return _efficientnet(
        inverted_residual_setting,
        0.5,
        last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )


def _efficientnet_b7(key, n_classes=1000, dtype: Any = None):
    inverted_residual_setting, last_channel = _efficientnet_conf(
        "efficientnet_b7", width_mult=2.0, depth_mult=3.1
    )
    return _efficientnet(
        inverted_residual_setting,
        0.5,
        last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )


def _efficientnet_v2_s(key, n_classes=1000, dtype: Any = None):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_v2_s")
    return _efficientnet(
        inverted_residual_setting,
        0.2,
        last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )


def _efficientnet_v2_m(key, n_classes=1000, dtype: Any = None):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_v2_m")
    return _efficientnet(
        inverted_residual_setting,
        0.3,
        last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )


def _efficientnet_v2_l(key, n_classes=1000, dtype: Any = None):
    inverted_residual_setting, last_channel = _efficientnet_conf("efficientnet_v2_l")
    return _efficientnet(
        inverted_residual_setting,
        0.4,
        last_channel,
        n_classes=n_classes,
        key=key,
        dtype=dtype,
    )


def _assert_model_and_weights_fit(
    model: Literal[
        "efficientnet_b0",
        "efficientnet_b1",
        "efficientnet_b2",
        "efficientnet_b3",
        "efficientnet_b4",
        "efficientnet_b5",
        "efficientnet_b6",
        "efficientnet_b7",
        "efficientnet_v2_s",
        "efficientnet_v2_m",
        "efficientnet_v2_l",
    ],
    weights: Literal[
        "efficientnet_b0_IMAGENET1K_V1",
        "efficientnet_b1_IMAGENET1K_V1",
        "efficientnet_b1_IMAGENET1K_V2",
        "efficientnet_b2_IMAGENET1K_V1",
        "efficientnet_b3_IMAGENET1K_V1",
        "efficientnet_b4_IMAGENET1K_V1",
        "efficientnet_b5_IMAGENET1K_V1",
        "efficientnet_b6_IMAGENET1K_V1",
        "efficientnet_b7_IMAGENET1K_V1",
        "efficientnet_v2_s_IMAGENET1K_V1",
        "efficientnet_v2_m_IMAGENET1K_V1",
        "efficientnet_v2_l_IMAGENET1K_V1",
    ],
):
    if "_" in weights:
        weights_model = weights.split("_IMAGENET")[0].lower()

        # Check if the specified weights are compatible with the model
        if weights_model != model:
            raise ValueError(
                f"Model {model} is incompatible with weights {weights}. "
                f"Expected weights starting with '{model}'."
            )


def load_efficientnet(
    model: Literal[
        "efficientnet_b0",
        "efficientnet_b1",
        "efficientnet_b2",
        "efficientnet_b3",
        "efficientnet_b4",
        "efficientnet_b5",
        "efficientnet_b6",
        "efficientnet_b7",
        "efficientnet_v2_s",
        "efficientnet_v2_m",
        "efficientnet_v2_l",
    ],
    weights: Literal[
        "efficientnet_b0_IMAGENET1K_V1",
        "efficientnet_b1_IMAGENET1K_V1",
        "efficientnet_b1_IMAGENET1K_V2",
        "efficientnet_b2_IMAGENET1K_V1",
        "efficientnet_b3_IMAGENET1K_V1",
        "efficientnet_b4_IMAGENET1K_V1",
        "efficientnet_b5_IMAGENET1K_V1",
        "efficientnet_b6_IMAGENET1K_V1",
        "efficientnet_b7_IMAGENET1K_V1",
        "efficientnet_v2_s_IMAGENET1K_V1",
        "efficientnet_v2_m_IMAGENET1K_V1",
        "efficientnet_v2_l_IMAGENET1K_V1",
    ]
    | None = None,
    n_classes: int = 1000,
    cache: bool = True,
    *,
    key: PRNGKeyArray | None = None,
    dtype: Any | None = None,  # Added optional dtype parameter
) -> tuple[EfficientNet, eqx.nn.State]:
    """
    Load an EfficientNet model with optional pre-trained weights.

    Args:
        model: The name of the EfficientNet model to load
        weights: The name of the weights to load (from _MODELS), or None for random init
        n_classes: Number of output classes
        cache: Whether to cache the model and weights.
        key: Random key for initialization. Defaults to jax.random.key(42).
        dtype: The data type for the model's parameters (e.g., jnp.float32).
               Defaults to jaxonmodels.functions.default_floating_dtype().

    Returns:
        A tuple of (model, state)
    """
    if key is None:
        key = jax.random.key(42)

    # Determine default dtype if not provided before passing to constructors
    if dtype is None:
        dtype = default_floating_dtype()
    assert dtype is not None  # Ensure dtype is set

    efficientnet, state = None, None  # Initialize

    match model:
        case "efficientnet_b0":
            efficientnet, state = _efficientnet_b0(key, n_classes, dtype=dtype)
        case "efficientnet_b1":
            efficientnet, state = _efficientnet_b1(key, n_classes, dtype=dtype)
        case "efficientnet_b2":
            efficientnet, state = _efficientnet_b2(key, n_classes, dtype=dtype)
        case "efficientnet_b3":
            efficientnet, state = _efficientnet_b3(key, n_classes, dtype=dtype)
        case "efficientnet_b4":
            efficientnet, state = _efficientnet_b4(key, n_classes, dtype=dtype)
        case "efficientnet_b5":
            efficientnet, state = _efficientnet_b5(key, n_classes, dtype=dtype)
        case "efficientnet_b6":
            efficientnet, state = _efficientnet_b6(key, n_classes, dtype=dtype)
        case "efficientnet_b7":
            efficientnet, state = _efficientnet_b7(key, n_classes, dtype=dtype)
        case "efficientnet_v2_s":
            efficientnet, state = _efficientnet_v2_s(key, n_classes, dtype=dtype)
        case "efficientnet_v2_m":
            efficientnet, state = _efficientnet_v2_m(key, n_classes, dtype=dtype)
        case "efficientnet_v2_l":
            efficientnet, state = _efficientnet_v2_l(key, n_classes, dtype=dtype)
        case _:
            raise ValueError(f"Unknown model name: {model}")

    if weights:
        _assert_model_and_weights_fit(model, weights)
        efficientnet, state = _with_weights(
            (efficientnet, state), weights, cache, dtype=dtype
        )

    assert efficientnet is not None
    assert state is not None
    return efficientnet, state



from .text import (
    SimpleTokenizer,
    clip_tokenize,
)
from .utils import (
    default_floating_dtype,
    dtype_to_str,
    make_divisible,
    make_ntuple,
    patch_merging_pad,
)

__all__ = [
    "SimpleTokenizer",
    "clip_tokenize",
    "make_divisible",
    "make_ntuple",
    "default_floating_dtype",
    "dtype_to_str",
    "patch_merging_pad",
]



import functools

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytest
import torch
from jaxonlayers.layers import BatchNorm, ConvNormActivation
from jaxonlayers.layers.attention import SqueezeExcitation
from statedict2pytree import (
    convert,
    move_running_fields_to_the_end,
    pytree_to_fields,
    state_dict_to_fields,
)
from torch.nn.modules import BatchNorm2d
from torchvision.models.efficientnet import MBConv as TorchMBConv
from torchvision.models.efficientnet import MBConvConfig as TorchMBConvConfig
from torchvision.ops.misc import Conv2dNormActivation as TorchConv2dNormActivation
from torchvision.ops.misc import SqueezeExcitation as TorchSqueezeExcitation

from jaxonmodels.functions.utils import default_floating_dtype
from jaxonmodels.models.efficientnet import (
    MBConv,
    MBConvConfig,
)


@pytest.mark.parametrize(
    "in_channels, out_channels, kernel_size, stride, padding, norm_layer, activation_layer",  # noqa
    [
        (3, 32, 3, 2, None, BatchNorm, jax.nn.silu),  # Base case
        (16, 24, 3, 1, None, BatchNorm, jax.nn.silu),  # Different channels
        (24, 40, 5, 2, 2, BatchNorm, jax.nn.silu),  # Larger kernel
        (40, 80, 3, 2, 1, BatchNorm, jax.nn.relu),  # Different activation
        (80, 112, 1, 1, 0, BatchNorm, jax.nn.silu),  # 1x1 conv
    ],
)
def test_Conv2dNormActivation_parametrized(
    in_channels,
    out_channels,
    kernel_size,
    stride,
    padding,
    norm_layer,
    activation_layer,
):
    dtype = default_floating_dtype()
    key = jax.random.key(42)

    # Create JAX version
    jax_conv2d_norm_act, state = eqx.nn.make_with_state(ConvNormActivation)(
        2,
        in_channels,
        out_channels,
        kernel_size=kernel_size,
        stride=stride,
        padding=padding,
        norm_layer=functools.partial(
            norm_layer, axis_name="batch", size=out_channels, dtype=dtype
        ),
        activation_layer=activation_layer,
        key=key,
        dtype=dtype,
    )

    # Create PyTorch version - convert JAX functions to torch equivalents
    torch_activation = (
        torch.nn.SiLU if activation_layer == jax.nn.silu else torch.nn.ReLU
    )
    torch_norm = torch.nn.BatchNorm2d if norm_layer == BatchNorm else None

    torch_conv2d_norm_act = TorchConv2dNormActivation(
        in_channels,
        out_channels,
        kernel_size=kernel_size,
        stride=stride,
        padding=padding,
        norm_layer=torch_norm,
        activation_layer=torch_activation,
    )

    # Convert weights
    weights_dict = torch_conv2d_norm_act.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    jaxfields, state_indices = pytree_to_fields((jax_conv2d_norm_act, state))

    jax_conv2d_norm_act, state = convert(
        weights_dict,
        (jax_conv2d_norm_act, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    jax_conv2d_norm_act, state = eqx.nn.inference_mode((jax_conv2d_norm_act, state))

    # Generate random input
    np.random.seed(42)
    input_height, input_width = 64, 64  # Example dimensions
    x = np.array(
        np.random.normal(size=(4, in_channels, input_height, input_width)),
        dtype=np.float32,
    )

    # Prepare models for inference
    torch_conv2d_norm_act.eval()
    key, subkey = jax.random.split(key)
    jax_conv2d_norm_act_pt = functools.partial(jax_conv2d_norm_act, key=subkey)

    # Run inference
    out_torch = torch_conv2d_norm_act.forward(torch.from_numpy(x))
    out_jax, state = eqx.filter_vmap(
        jax_conv2d_norm_act_pt, in_axes=(0, None), out_axes=(0, None), axis_name="batch"
    )(jnp.array(x), state)

    # Verify outputs match
    assert np.allclose(out_torch.detach().numpy(), np.array(out_jax), atol=1e-5)


@pytest.mark.parametrize(
    "in_channels, out_channels, kernel_size, stride, groups",
    [
        (32, 32, 3, 1, 32),  # Depthwise convolution
        (64, 64, 5, 2, 64),  # Depthwise with larger kernel and stride
    ],
)
def test_Conv2dNormActivation_groups(
    in_channels, out_channels, kernel_size, stride, groups
):
    dtype = default_floating_dtype()
    key = jax.random.key(42)

    # Create JAX version
    jax_conv2d_norm_act, state = eqx.nn.make_with_state(ConvNormActivation)(
        2,
        in_channels,
        out_channels,
        kernel_size=kernel_size,
        stride=stride,
        groups=groups,
        norm_layer=functools.partial(
            BatchNorm, axis_name="batch", size=out_channels, dtype=dtype
        ),
        activation_layer=jax.nn.silu,
        key=key,
        dtype=dtype,
    )

    # Create PyTorch version
    torch_conv2d_norm_act = TorchConv2dNormActivation(
        in_channels,
        out_channels,
        kernel_size=kernel_size,
        stride=stride,
        groups=groups,
        norm_layer=torch.nn.BatchNorm2d,
        activation_layer=torch.nn.SiLU,
    )

    # Convert weights
    weights_dict = torch_conv2d_norm_act.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    jaxfields, state_indices = pytree_to_fields((jax_conv2d_norm_act, state))

    jax_conv2d_norm_act, state = convert(
        weights_dict,
        (jax_conv2d_norm_act, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    jax_conv2d_norm_act, state = eqx.nn.inference_mode((jax_conv2d_norm_act, state))

    # Generate random input
    np.random.seed(42)
    input_height, input_width = 32, 32  # Example dimensions
    x = np.array(
        np.random.normal(size=(2, in_channels, input_height, input_width)),
        dtype=np.float32,
    )

    # Prepare models for inference
    torch_conv2d_norm_act.eval()
    key, subkey = jax.random.split(key)
    jax_conv2d_norm_act_pt = functools.partial(jax_conv2d_norm_act, key=subkey)

    # Run inference
    out_torch = torch_conv2d_norm_act.forward(torch.from_numpy(x))
    out_jax, state = eqx.filter_vmap(
        jax_conv2d_norm_act_pt, in_axes=(0, None), out_axes=(0, None), axis_name="batch"
    )(jnp.array(x), state)

    # Verify outputs match
    assert np.allclose(out_torch.detach().numpy(), np.array(out_jax), atol=1e-6)


@pytest.mark.parametrize(
    "input_channels, squeeze_factor",
    [
        (96, 4),  # Original test case
        (16, 4),  # Smaller input channels
        (192, 8),  # Larger input channels, different squeeze factor
        (24, 2),  # Different combination
        (80, 16),  # Different combination
    ],
)
def test_SqueezeExcitation_parametrized(input_channels, squeeze_factor):
    squeeze_channels = max(1, input_channels // squeeze_factor)

    key = jax.random.key(42)
    jax_squeeze = SqueezeExcitation(input_channels, squeeze_channels, key=key)
    torch_squeeze = TorchSqueezeExcitation(input_channels, squeeze_channels)

    weights_dict = torch_squeeze.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    jaxfields, state_indices = pytree_to_fields(jax_squeeze)

    jax_squeeze = convert(
        weights_dict,
        jax_squeeze,
        jaxfields,
        state_indices,
        torchfields,
    )

    # Generate input with appropriate shape for testing
    np.random.seed(42)
    height, width = 28, 28  # Example dimensions
    x = np.array(
        np.random.normal(size=(input_channels, height, width)), dtype=np.float32
    )

    jax_output = jax_squeeze(jnp.array(x))
    torch_output = torch_squeeze(torch.from_numpy(x))

    assert np.allclose(np.array(jax_output), torch_output.detach().numpy(), atol=1e-6)


@pytest.mark.parametrize(
    "expand_ratio, kernel, stride, input_channels, out_channels, num_layers, stochastic_depth_prob, height, width",  # noqa
    [
        (1.0, 3, 1, 32, 16, 1, 0.0, 112, 112),
        (6.0, 3, 2, 16, 24, 2, 0.0125, 112, 112),
        (6.0, 3, 1, 24, 24, 2, 0.025, 56, 56),
        (6.0, 5, 2, 24, 40, 2, 0.0375, 56, 56),
        (6.0, 5, 1, 40, 40, 2, 0.05, 28, 28),
        (6.0, 3, 2, 40, 80, 3, 0.0625, 28, 28),
        (6.0, 3, 1, 80, 80, 3, 0.075, 14, 14),
        (6.0, 3, 1, 80, 80, 3, 0.0875, 14, 14),
        (6.0, 5, 1, 80, 112, 3, 0.1, 14, 14),
        (6.0, 5, 1, 112, 112, 3, 0.1125, 14, 14),
        (6.0, 5, 1, 112, 112, 3, 0.125, 14, 14),
        (6.0, 5, 2, 112, 192, 4, 0.1375, 14, 14),
        (6.0, 5, 1, 192, 192, 4, 0.15, 7, 7),
        (6.0, 5, 1, 192, 192, 4, 0.1625, 7, 7),
        (6.0, 5, 1, 192, 192, 4, 0.175, 7, 7),
        (6.0, 3, 1, 192, 320, 1, 0.1875, 7, 7),
    ],
)
def test_mbconv_forward(
    expand_ratio,
    kernel,
    stride,
    input_channels,
    out_channels,
    num_layers,
    stochastic_depth_prob,
    height,
    width,
):
    dtype = default_floating_dtype()
    t_cnf = TorchMBConvConfig(
        expand_ratio=expand_ratio,
        kernel=kernel,
        stride=stride,
        input_channels=input_channels,
        out_channels=out_channels,
        num_layers=num_layers,
        block=TorchMBConv,
    )

    torch_mbconv = TorchMBConv(t_cnf, stochastic_depth_prob, BatchNorm2d)
    j_cnf = MBConvConfig(
        expand_ratio=expand_ratio,
        kernel=kernel,
        stride=stride,
        input_channels=input_channels,
        out_channels=out_channels,
        num_layers=num_layers,
    )
    key = jax.random.key(42)
    jax_mbconv, state = eqx.nn.make_with_state(MBConv)(
        j_cnf, stochastic_depth_prob, key=key, dtype=dtype
    )

    weights_dict = torch_mbconv.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    jaxfields, state_indices = pytree_to_fields((jax_mbconv, state))

    jax_mbconv, state = convert(
        weights_dict,
        (jax_mbconv, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    jax_mbconv, state = eqx.nn.inference_mode((jax_mbconv, state))

    np.random.seed(42)
    x = np.array(
        np.random.normal(size=(2, input_channels, height, width)), dtype=np.float32
    )
    key, subkey = jax.random.split(key)
    jax_mbconv_pt = functools.partial(jax_mbconv, key=subkey)
    torch_mbconv.eval()

    jax_output, state = eqx.filter_vmap(
        jax_mbconv_pt, in_axes=(0, None), out_axes=(0, None), axis_name="batch"
    )(jnp.array(x), state)

    torch_output = torch_mbconv(torch.from_numpy(x))

    # print(jax_output[0])
    # print(torch_output[0])

    assert np.allclose(np.array(jax_output), torch_output.detach().numpy(), atol=1e-5)



import os
from pathlib import Path
from urllib.request import urlretrieve

import equinox as eqx
import jax
import jax.numpy as jnp
from beartype.typing import Any
from jaxonlayers.layers import LocalResponseNormalization
from jaxtyping import Array, Float, PRNGKeyArray
from statedict2pytree.converter import (
    convert,
    pytree_to_fields,
    state_dict_to_fields,
)

from jaxonmodels.functions import default_floating_dtype, dtype_to_str


class AlexNet(eqx.Module):
    conv1: eqx.nn.Conv2d
    conv2: eqx.nn.Conv2d
    conv3: eqx.nn.Conv2d
    conv4: eqx.nn.Conv2d
    conv5: eqx.nn.Conv2d
    lrn1: LocalResponseNormalization
    lrn2: LocalResponseNormalization
    max_pool_1: eqx.nn.MaxPool2d
    max_pool_2: eqx.nn.MaxPool2d
    max_pool_3: eqx.nn.MaxPool2d

    dense1: eqx.nn.Linear
    dense2: eqx.nn.Linear

    dropout1: eqx.nn.Dropout
    dropout2: eqx.nn.Dropout

    final: eqx.nn.Linear

    inference: bool

    def __init__(
        self,
        *,
        n_classes: int,
        key: PRNGKeyArray,
        inference: bool = False,
        dtype: Any | None = None,
    ):
        self.inference = inference
        _, *subkeys = jax.random.split(key, 10)
        self.conv1 = eqx.nn.Conv2d(
            in_channels=3,
            out_channels=64,
            kernel_size=11,
            stride=4,
            padding=2,
            key=subkeys[0],
            dtype=dtype,
        )
        self.conv2 = eqx.nn.Conv2d(
            in_channels=64,
            out_channels=192,
            kernel_size=5,
            stride=1,
            padding=2,
            key=subkeys[1],
            dtype=dtype,
        )
        self.conv3 = eqx.nn.Conv2d(
            in_channels=192,
            out_channels=384,
            kernel_size=3,
            stride=1,
            padding=1,
            key=subkeys[2],
            dtype=dtype,
        )
        self.conv4 = eqx.nn.Conv2d(
            in_channels=384,
            out_channels=256,
            kernel_size=3,
            stride=1,
            padding=1,
            key=subkeys[3],
            dtype=dtype,
        )
        self.conv5 = eqx.nn.Conv2d(
            in_channels=256,
            out_channels=256,
            kernel_size=3,
            stride=1,
            padding=1,
            key=subkeys[4],
            dtype=dtype,
        )

        self.lrn1 = LocalResponseNormalization()
        self.lrn2 = LocalResponseNormalization()
        self.max_pool_1 = eqx.nn.MaxPool2d(kernel_size=3, stride=2)
        self.max_pool_2 = eqx.nn.MaxPool2d(kernel_size=3, stride=2)
        self.max_pool_3 = eqx.nn.MaxPool2d(kernel_size=3, stride=2)

        self.dense1 = eqx.nn.Linear(
            in_features=9216, out_features=4096, key=subkeys[5], dtype=dtype
        )
        self.dense2 = eqx.nn.Linear(
            in_features=4096, out_features=4096, key=subkeys[6], dtype=dtype
        )

        self.dropout1 = eqx.nn.Dropout()
        self.dropout2 = eqx.nn.Dropout()

        self.final = eqx.nn.Linear(
            in_features=4096, out_features=n_classes, key=subkeys[7], dtype=dtype
        )

    def __call__(
        self,
        x: Float[Array, "c h w"],
        key: PRNGKeyArray | None = None,
    ) -> Array:
        if self.inference:
            key, subkey = None, None
        else:
            if not key:
                raise ValueError("Expected a random key in training mode, got None")
            else:
                key, subkey = jax.random.split(key)
        x = self.conv1(x)
        x = jax.nn.relu(x)
        x = self.lrn1(x)
        x = self.max_pool_1(x)
        x = self.conv2(x)
        x = jax.nn.relu(x)
        x = self.lrn2(x)
        x = self.max_pool_2(x)

        x = self.conv3(x)
        x = jax.nn.relu(x)
        x = self.conv4(x)
        x = jax.nn.relu(x)
        x = self.conv5(x)
        x = jax.nn.relu(x)

        x = self.max_pool_3(x)
        x = jnp.ravel(x)

        x = self.dense1(x)
        x = jax.nn.relu(x)
        x = self.dropout1(x, key=key)
        x = self.dense2(x)
        x = jax.nn.relu(x)
        x = self.dropout2(x, key=subkey)

        x = self.final(x)
        return x


def alexnet(
    with_weights: bool = False, cache: bool = True, dtype: Any | None = None
) -> "AlexNet":
    if dtype is None:
        dtype = default_floating_dtype()
    assert dtype is not None
    dtype_str = dtype_to_str(dtype)
    jaxonmodels_dir = os.path.expanduser("~/.jaxonmodels/models")
    os.makedirs(jaxonmodels_dir, exist_ok=True)
    alexnet = AlexNet(n_classes=1000, key=jax.random.key(0), dtype=dtype)
    if not with_weights:
        return alexnet
    if cache:
        if os.path.exists(str(Path(jaxonmodels_dir) / f"alexnet-{dtype_str}.eqx")):
            return eqx.tree_deserialise_leaves(
                str(Path(jaxonmodels_dir) / f"alexnet-{dtype_str}.eqx"), alexnet
            )

    weights_url = "https://download.pytorch.org/models/alexnet-owt-7be5be79.pth"
    weights_dir = os.path.expanduser("~/.jaxonmodels/pytorch_weights")
    os.makedirs(weights_dir, exist_ok=True)

    weights_file = os.path.join(weights_dir, "alexnet-owt-7be5be79.pth")
    if not os.path.exists(weights_file):
        urlretrieve(weights_url, weights_file)

    import torch

    weights_dict = torch.load(
        weights_file, map_location=torch.device("cpu"), weights_only=True
    )

    torchfields = state_dict_to_fields(weights_dict)
    jaxfields, _ = pytree_to_fields(alexnet)

    alexnet = convert(weights_dict, alexnet, jaxfields, None, torchfields, dtype=dtype)

    if cache:
        eqx.tree_serialise_leaves(
            str(Path(jaxonmodels_dir) / f"alexnet-{dtype_str}.eqx"), alexnet
        )

    return alexnet



import functools

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytest
import torch
from jaxonlayers.layers import LayerNorm
from statedict2pytree import (
    convert,
    move_running_fields_to_the_end,
    pytree_to_fields,
    state_dict_to_fields,
)
from torchvision.models.swin_transformer import PatchMerging as TorchPatchMerging
from torchvision.models.swin_transformer import PatchMergingV2 as TorchPatchMergingV2
from torchvision.models.swin_transformer import (
    ShiftedWindowAttention as TorchShiftedWindowAttentionV1,
)
from torchvision.models.swin_transformer import (
    ShiftedWindowAttentionV2 as TorchShiftedWindowAttentionV2,
)
from torchvision.models.swin_transformer import SwinTransformer as TorchSwinTransformer
from torchvision.models.swin_transformer import (
    SwinTransformerBlock as TorchSwinTransformerBlock,
)
from torchvision.models.swin_transformer import (
    SwinTransformerBlockV2 as TorchSwinTransformerBlockV2,
)

from jaxonmodels.functions import default_floating_dtype
from jaxonmodels.models.swin_transformer import (
    PatchMerging,
    PatchMergingV2,
    ShiftedWindowAttention,
    ShiftedWindowAttentionV2,
    SwinTransformer,
    SwinTransformerBlock,
    SwinTransformerBlockV2,
)
from jaxonmodels.statedict2pytree import model_orders


@pytest.mark.parametrize(
    "C, H, W",
    [
        (32, 55, 55),
        (32, 224, 224),
        (16, 56, 56),
        (32, 112, 56),
    ],
)
def test_patch_merging_layers(C, H, W):
    batch_size = 8
    dtype = default_floating_dtype()
    torch_patch = TorchPatchMerging(dim=C)
    jax_patch, state = eqx.nn.make_with_state(PatchMerging)(
        dim=C,
        norm_layer=functools.partial(LayerNorm, dtype=dtype),
        inference=False,
        dtype=dtype,
        key=jax.random.key(42),
    )

    jax_patch, state = eqx.nn.inference_mode((jax_patch, state))

    weights_dict = torch_patch.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    jaxfields, state_indices = pytree_to_fields((jax_patch, state))

    jax_patch, state = convert(
        weights_dict,
        (jax_patch, state),
        jaxfields,
        state_indices,
        torchfields,
    )
    np.random.seed(42)
    x = np.array(np.random.normal(size=(batch_size, H, W, C)), dtype=np.float32)

    t_out = torch_patch(torch.from_numpy(x))
    j_out, state = eqx.filter_vmap(jax_patch, in_axes=(0, None), out_axes=(0, None))(
        jnp.array(x), state
    )

    assert np.allclose(np.array(j_out), t_out.detach().numpy(), atol=1e-6)


@pytest.mark.parametrize(
    "C, window_size, shift_size, num_heads, qkv_bias, proj_bias, attention_dropout, dropout, H, W",  # noqa
    [
        (96, [7, 7], [0, 0], 3, True, True, 0.0, 0.0, 56, 56),
        (96, [7, 7], [3, 3], 3, True, True, 0.0, 0.0, 56, 56),
        (192, [7, 7], [0, 0], 6, True, True, 0.0, 0.0, 56, 56),
        (192, [7, 7], [3, 3], 6, True, True, 0.0, 0.0, 56, 56),
        (384, [7, 7], [0, 0], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [7, 7], [3, 3], 12, True, True, 0.0, 0.0, 56, 56),
        (768, [7, 7], [0, 0], 24, True, True, 0.0, 0.0, 56, 56),
        (768, [7, 7], [3, 3], 24, True, True, 0.0, 0.0, 56, 56),
        (96, [8, 8], [0, 0], 3, True, True, 0.0, 0.0, 56, 56),
        (96, [8, 8], [4, 4], 3, True, True, 0.0, 0.0, 56, 56),
        (192, [8, 8], [0, 0], 6, True, True, 0.0, 0.0, 56, 56),
        (192, [8, 8], [4, 4], 6, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [0, 0], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [4, 4], 12, True, True, 0.0, 0.0, 56, 56),
        (768, [8, 8], [0, 0], 24, True, True, 0.0, 0.0, 56, 56),
        (768, [8, 8], [4, 4], 24, True, True, 0.0, 0.0, 56, 56),
    ],
)
def test_swin_attention_v1(
    C,
    window_size,
    shift_size,
    num_heads,
    qkv_bias,
    proj_bias,
    attention_dropout,
    dropout,
    H,
    W,
):
    batch_size = 8
    dtype = default_floating_dtype()
    torch_swin = TorchShiftedWindowAttentionV1(C, window_size, shift_size, num_heads)

    jax_swin, state = eqx.nn.make_with_state(ShiftedWindowAttention)(
        dim=C,
        shift_size=shift_size,
        num_heads=num_heads,
        qkv_bias=qkv_bias,
        window_size=window_size,
        proj_bias=proj_bias,
        attention_dropout=attention_dropout,
        dropout=dropout,
        inference=False,
        dtype=dtype,
        key=jax.random.key(42),
    )

    jax_swin, state = eqx.nn.inference_mode((jax_swin, state))

    weights_dict = torch_swin.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    torchfields = move_running_fields_to_the_end(
        torchfields, identifier="relative_position_index"
    )
    jaxfields, state_indices = pytree_to_fields((jax_swin, state))

    # for t, j in zip(torchfields, jaxfields):
    #     print(t.path, t.shape, jax.tree_util.keystr(j.path), j.shape)

    jax_swin, state = convert(
        weights_dict,
        (jax_swin, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    np.random.seed(42)
    x = np.array(np.random.normal(size=(batch_size, H, W, C)), dtype=np.float32)

    t_out = torch_swin(torch.from_numpy(x))

    j_out, state = eqx.filter_vmap(jax_swin, in_axes=(0, None), out_axes=(0, None))(
        jnp.array(x), state
    )

    assert np.allclose(np.array(j_out), t_out.detach().numpy(), atol=1e-5)


@pytest.mark.parametrize(
    "C, window_size, shift_size, num_heads, qkv_bias, proj_bias, attention_dropout, dropout, H, W",  # noqa
    [
        (96, [8, 8], [0, 0], 3, True, True, 0.0, 0.0, 56, 56),
        (96, [8, 8], [4, 4], 3, True, True, 0.0, 0.0, 56, 56),
        (192, [8, 8], [0, 0], 6, True, True, 0.0, 0.0, 56, 56),
        (192, [8, 8], [4, 4], 6, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [0, 0], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [4, 4], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [0, 0], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [4, 4], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [0, 0], 12, True, True, 0.0, 0.0, 56, 56),
        (384, [8, 8], [4, 4], 12, True, True, 0.0, 0.0, 56, 56),
        (768, [8, 8], [0, 0], 24, True, True, 0.0, 0.0, 56, 56),
        (768, [8, 8], [4, 4], 24, True, True, 0.0, 0.0, 56, 56),
    ],
)
def test_swin_attention_v2(
    C,
    window_size,
    shift_size,
    num_heads,
    qkv_bias,
    proj_bias,
    attention_dropout,
    dropout,
    H,
    W,
):
    batch_size = 8
    dtype = default_floating_dtype()
    torch_swin = TorchShiftedWindowAttentionV2(C, window_size, shift_size, num_heads)

    jax_swin, state = eqx.nn.make_with_state(ShiftedWindowAttentionV2)(
        dim=C,
        shift_size=shift_size,
        num_heads=num_heads,
        qkv_bias=qkv_bias,
        window_size=window_size,
        proj_bias=proj_bias,
        attention_dropout=attention_dropout,
        dropout=dropout,
        inference=False,
        dtype=dtype,
        key=jax.random.key(42),
    )

    jax_swin, state = eqx.nn.inference_mode((jax_swin, state))

    weights_dict = torch_swin.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    torchfields = move_running_fields_to_the_end(
        torchfields, identifier="relative_position_index"
    )
    torchfields = move_running_fields_to_the_end(
        torchfields, identifier="relative_coords_table"
    )

    jaxfields, state_indices = pytree_to_fields((jax_swin, state))

    for t, j in zip(torchfields, jaxfields):
        print(t.path, t.shape, jax.tree_util.keystr(j.path), j.shape)

    jax_swin, state = convert(
        weights_dict,
        (jax_swin, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    np.random.seed(42)
    x = np.array(np.random.normal(size=(batch_size, H, W, C)), dtype=np.float32)

    t_out = torch_swin(torch.from_numpy(x))

    j_out, state = eqx.filter_vmap(jax_swin, in_axes=(0, None), out_axes=(0, None))(
        jnp.array(x), state
    )

    assert np.allclose(np.array(j_out), t_out.detach().numpy(), atol=1e-5)


@pytest.mark.parametrize(
    "dim, num_heads, window_size, shift_size, mlp_ratio, qkv_bias, proj_bias, dropout, attention_dropout, stochastic_depth_prob, H, W",  # noqa
    [
        (96, 3, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 56, 56),
        (96, 3, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.01818, 56, 56),
        (192, 6, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0363636, 28, 28),
        (192, 6, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.0545456, 28, 28),
        (384, 12, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.072727, 14, 14),
        (384, 12, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.090909, 14, 14),
        (384, 12, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.109090, 14, 14),
        (384, 12, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.127272, 14, 14),
        (384, 12, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.145454, 14, 14),
        (384, 12, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.163636, 14, 14),
        (768, 24, [7, 7], [0, 0], 4.0, True, True, 0.0, 0.0, 0.181818, 7, 7),
        (768, 24, [7, 7], [3, 3], 4.0, True, True, 0.0, 0.0, 0.2, 7, 7),
    ],
)
def test_swin_transformer_block(
    dim,
    num_heads,
    window_size,
    shift_size,
    mlp_ratio,
    qkv_bias,
    proj_bias,
    dropout,
    attention_dropout,
    stochastic_depth_prob,
    H,
    W,
):
    batch_size = 2
    dtype = default_floating_dtype()
    key = jax.random.key(43)
    torch.manual_seed(42)
    np.random.seed(42)

    # --- PyTorch Model ---

    torch_block = TorchSwinTransformerBlock(
        dim=dim,
        num_heads=num_heads,
        window_size=window_size,
        shift_size=shift_size,
        mlp_ratio=mlp_ratio,
        dropout=dropout,
        attention_dropout=attention_dropout,
        stochastic_depth_prob=stochastic_depth_prob,
        norm_layer=torch.nn.LayerNorm,
        attn_layer=TorchShiftedWindowAttentionV1,
    )
    torch_block.eval()

    # --- JAX Model ---
    #  Always use the V1 attention module.
    jax_block, state = eqx.nn.make_with_state(SwinTransformerBlock)(
        dim=dim,
        num_heads=num_heads,
        window_size=window_size,
        shift_size=shift_size,
        mlp_ratio=mlp_ratio,
        dropout=dropout,
        attention_dropout=attention_dropout,
        stochastic_depth_prob=stochastic_depth_prob,
        norm_layer=functools.partial(eqx.nn.LayerNorm, eps=1e-05),  # pyright: ignore
        attn_layer=ShiftedWindowAttention,
        inference=False,
        dtype=dtype,
        key=key,
    )
    jax_block, state = eqx.nn.inference_mode((jax_block, state))

    # --- Weight Conversion ---
    weights_dict = torch_block.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    torchfields = move_running_fields_to_the_end(
        torchfields,
        identifier="relative_position_index",
    )

    jaxfields, state_indices = pytree_to_fields(
        (
            jax_block,
            state,
        )
    )

    jax_block, state = convert(
        weights_dict,
        (jax_block, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    # --- Inference and Comparison ---
    x_np = np.array(np.random.uniform(size=(batch_size, H, W, dim)), dtype=np.float32)
    x_torch = torch.from_numpy(x_np)
    x_jax = jnp.array(x_np)

    t_out = torch_block(x_torch)
    jax_block_pt = functools.partial(jax_block, key=jax.random.key(2100))
    j_out, state = eqx.filter_vmap(jax_block_pt, in_axes=(0, None), out_axes=(0, None))(
        x_jax, state
    )

    assert np.allclose(np.array(j_out), t_out.detach().numpy(), atol=1e-3)


@pytest.mark.parametrize(
    "dim, num_heads, window_size, shift_size, mlp_ratio, qkv_bias, proj_bias, dropout, attention_dropout, stochastic_depth_prob, H, W",  # noqa
    [
        (96, 3, [8, 8], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0, 56, 56),
        (96, 3, [8, 8], [4, 4], 4.0, True, True, 0.0, 0.0, 0.01818, 56, 56),
        (192, 6, [8, 8], [0, 0], 4.0, True, True, 0.0, 0.0, 0.0363636, 28, 28),
        (192, 6, [8, 8], [4, 4], 4.0, True, True, 0.0, 0.0, 0.0545456, 28, 28),
        (384, 12, [8, 8], [0, 0], 4.0, True, True, 0.0, 0.0, 0.072727, 14, 14),
        (384, 12, [8, 8], [4, 4], 4.0, True, True, 0.0, 0.0, 0.090909, 14, 14),
        (768, 24, [8, 8], [0, 0], 4.0, True, True, 0.0, 0.0, 0.181818, 7, 7),
        (768, 24, [8, 8], [4, 4], 4.0, True, True, 0.0, 0.0, 0.2, 7, 7),
    ],
)
def test_swin_transformer_blockv2(
    dim,
    num_heads,
    window_size,
    shift_size,
    mlp_ratio,
    qkv_bias,
    proj_bias,
    dropout,
    attention_dropout,
    stochastic_depth_prob,
    H,
    W,
):
    batch_size = 2
    dtype = default_floating_dtype()
    key = jax.random.key(43)
    torch.manual_seed(42)
    np.random.seed(42)

    # --- PyTorch Model ---
    torch_block = TorchSwinTransformerBlockV2(
        dim=dim,
        num_heads=num_heads,
        window_size=window_size,
        shift_size=shift_size,
        mlp_ratio=mlp_ratio,
        dropout=dropout,
        attention_dropout=attention_dropout,
        stochastic_depth_prob=stochastic_depth_prob,
        norm_layer=torch.nn.LayerNorm,
        attn_layer=TorchShiftedWindowAttentionV2,
    )
    torch_block.eval()

    # --- JAX Model ---
    jax_block, state = eqx.nn.make_with_state(SwinTransformerBlockV2)(
        dim=dim,
        num_heads=num_heads,
        window_size=window_size,
        shift_size=shift_size,
        mlp_ratio=mlp_ratio,
        dropout=dropout,
        attention_dropout=attention_dropout,
        stochastic_depth_prob=stochastic_depth_prob,
        norm_layer=functools.partial(eqx.nn.LayerNorm, eps=1e-05),  # pyright: ignore
        attn_layer=ShiftedWindowAttentionV2,
        inference=False,
        dtype=dtype,
        key=key,
    )
    jax_block, state = eqx.nn.inference_mode((jax_block, state))

    # --- Weight Conversion ---
    weights_dict = torch_block.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    torchfields = move_running_fields_to_the_end(
        torchfields,
        identifier="relative_position_index",
    )
    torchfields = move_running_fields_to_the_end(
        torchfields,
        identifier="relative_coords_table",
    )

    jaxfields, state_indices = pytree_to_fields(
        (
            jax_block,
            state,
        )
    )

    jax_block, state = convert(
        weights_dict,
        (jax_block, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    # --- Inference and Comparison ---
    x_np = np.array(np.random.uniform(size=(batch_size, H, W, dim)), dtype=np.float32)
    x_torch = torch.from_numpy(x_np)
    x_jax = jnp.array(x_np)

    t_out = torch_block(x_torch)
    jax_block_pt = functools.partial(jax_block, key=jax.random.key(2100))
    j_out, state = eqx.filter_vmap(jax_block_pt, in_axes=(0, None), out_axes=(0, None))(
        x_jax, state
    )

    assert np.allclose(np.array(j_out), t_out.detach().numpy(), atol=1e-3)


def test_swinv1():
    patch_size = [4, 4]
    embed_dim = 96
    depths = [2, 2, 6, 2]
    num_heads = [3, 6, 12, 24]
    window_size = [7, 7]
    mlp_ratio = 4.0
    dropout = 0.0
    attention_dropout = 0.0
    stochastic_depth_prob = 0.2
    num_classes = 1000

    torch_swin = TorchSwinTransformer(
        patch_size,
        embed_dim,
        depths,
        num_heads,
        window_size,
        mlp_ratio,
        dropout,
        attention_dropout,
        stochastic_depth_prob,
        num_classes,
        block=TorchSwinTransformerBlock,
    )
    torch_swin.eval()
    np.random.seed(42)

    x = np.array(np.random.normal(size=(1, 3, 224, 224)), dtype=np.float32)

    x_t = torch.Tensor(x)

    jax_swin, state = eqx.nn.make_with_state(SwinTransformer)(
        patch_size,
        embed_dim,
        depths,
        num_heads,
        window_size,
        mlp_ratio,
        dropout,
        attention_dropout,
        stochastic_depth_prob,
        num_classes,
        norm_layer=None,
        block=SwinTransformerBlock,
        downsample_layer=PatchMerging,
        attn_layer=ShiftedWindowAttention,
        inference=True,
        key=jax.random.key(42),
    )

    weights_dict = torch_swin.state_dict()
    torchfields = state_dict_to_fields(weights_dict)

    jaxfields, state_indices = pytree_to_fields(
        (
            jax_swin,
            state,
        ),
        model_order=model_orders.get_swin_model_order(1),
    )
    jax_swin, state = convert(
        weights_dict,
        (jax_swin, state),
        jaxfields,
        state_indices,
        torchfields,
    )

    jax_swin, state = eqx.nn.inference_mode((jax_swin, state))
    key = jax.random.key(22)
    jax_swin_pt = functools.partial(jax_swin, key=key)
    jax_out, state = eqx.filter_vmap(
        jax_swin_pt, in_axes=(0, None), out_axes=(0, None), axis_name="batch"
    )(jnp.array(x), state)

    t_out = torch_swin.forward(x_t).detach().numpy()

    jax_out = np.array(jax_out)

    assert np.allclose(jax_out, t_out, atol=1e-3)


def test_swinv2():
    patch_size = [4, 4]
    embed_dim = 96
    depths = [2, 2, 6, 2]
    num_heads = [3, 6, 12, 24]
    window_size = [7, 7]
    mlp_ratio = 4.0
    dropout = 0.0
    attention_dropout = 0.0
    stochastic_depth_prob = 0.2
    num_classes = 1000

    torch_swin = TorchSwinTransformer(
        patch_size,
        embed_dim,
        depths,
        num_heads,
        window_size,
        mlp_ratio,
        dropout,
        attention_dropout,
        stochastic_depth_prob,
        num_classes,
        downsample_layer=TorchPatchMergingV2,
        block=TorchSwinTransformerBlockV2,
    )
    torch_swin.eval()
    np.random.seed(42)

    x = np.array(np.random.normal(size=(1, 3, 224, 224)), dtype=np.float32)

    x_t = torch.Tensor(x)

    jax_swin, state = eqx.nn.make_with_state(SwinTransformer)(
        patch_size,
        embed_dim,
        depths,
        num_heads,
        window_size,
        mlp_ratio,
        dropout,
        attention_dropout,
        stochastic_depth_prob,
        num_classes,
        norm_layer=None,
        block=SwinTransformerBlockV2,
        downsample_layer=PatchMergingV2,
        attn_layer=ShiftedWindowAttentionV2,
        inference=True,
        key=jax.random.key(42),
    )

    weights_dict = torch_swin.state_dict()
    torchfields = state_dict_to_fields(weights_dict)
    jaxfields, state_indices = pytree_to_fields(
        (
            jax_swin,
            state,
        ),
        model_order=model_orders.get_swin_model_order(2),
    )

    # Print zipped jaxfields and torchfields for debugging
    for t, j in zip(torchfields, jaxfields):
        print(t.path, t.shape, jax.tree_util.keystr(j.path), j.shape)

    jax_swin, state = convert(
        weights_dict,
        (jax_swin, state),
        jaxfields,
        state_indices,
        torchfields,
    )
    jax_swin, state = eqx.nn.inference_mode((jax_swin, state))

    key = jax.random.key(22)
    jax_swin_pt = functools.partial(jax_swin, key=key)
    jax_out, state = eqx.filter_vmap(
        jax_swin_pt, in_axes=(0, None), out_axes=(0, None), axis_name="batch"
    )(jnp.array(x), state)

    t_out = torch_swin.forward(x_t).detach().numpy()

    jax_out = np.array(jax_out)

    assert np.allclose(jax_out, t_out, atol=1e-3)
